using namespace std;
/*************************************************************** 
* 
* DalAttributable.cc          & XMM Science Analysis System (SAS) 
* 
* Class representing a set of attributes.
* 
* (c) 1997 ESA - Design by Marco Beijersbergen
*                Written by Mark Thomas, CARA, and Marco Beijersbergen, CRI
* 
* $Id: DalAttributable.cc,v 1.2 2004/03/05 14:43:00 eojero Exp $
* 
*****************************************************************/ 

#include <string>
#include <list>

#include <errstr.h>
#include <Tracer.h>
#include <SimpleScanner.h>

#include "DalAttributable.h"
#include "DalAttribute.h"


DalAttributable::DalAttributable(
		const string &		name,
		const string &		label,
		AccessMode		accessMode,
		const DalDal *		parentAccessMode )
	:
		DalLabelled( name, label ),
		DalDal( accessMode, parentAccessMode )
{
	PROC
	ENDPROC
}


DalAttributable::~DalAttributable()
{
	PROC

	for( AttributeList::iterator i = _attributes.begin();
			i != _attributes.end(); ++i)
		delete (*i);

	ENDPROC
}


Attributable& DalAttributable::operator=( const Attributable& org )
{
	PROC

	// **************************
	// Delete the old attributes.
	// **************************
	for(
		AttributeList::iterator i = _attributes.begin();
		i != _attributes.end();
		++i )
	{
		delete (*i);

	} // for

	_attributes.erase(_attributes.begin(),_attributes.end());


	// **********************************
	// Keep the old comments and history.
	// **********************************
	// _comments.erase(_comments.begin(),_comments.end());
	// _history.erase(_history.begin(),_history.end());


	// **********************************************
	// Add the new attributes, history, and comments.
	// **********************************************
	addAttributes( &org );


	RETURN *this;
}


void
DalAttributable::checkValidAttributeNumber( unsigned int attributeNumber ) const
{
	PROC
	// ***************************************************
	// Check that the requested attribute number is valid.
	// ***************************************************
	if( attributeNumber >= attributes() )
	{
		errstr << "There is no attribute with number " << attributeNumber << " in data set " << name() << "; it has " << attributes() << " attributes" << Sas::error("attributeNumberNotFound");

	} // if

	ENDPROC
}


Attribute *
DalAttributable::addAttribute( const Attribute* org )
{
	PROC

	checkIfReadOnly();

	const string& name = org->name();

	if( hasAttribute( name ) )
	{
		Attribute * attribute;

		attribute = this -> attribute( name );
		*attribute = *org;

		ALTRETURNV( attribute );
	}
	else
	{
		DalAttribute* attribute;

		attribute = new DalAttribute( org, this );
		_attributes.push_back( attribute );

		ALTRETURNV( attribute );
	} // if
}


Attribute *
DalAttributable::addAttribute(
				const string &		name,
				const string &		comment,
				const string &		units)
{
	PROC

	checkIfReadOnly();

	const string	attributeName = DalAttributable::parseName( name );

	if ( hasAttribute( attributeName ) )
	{
		Attribute * attribute;

		attribute = this->attribute( attributeName );
		attribute -> units(units);
		attribute -> relabel(comment);

		ALTRETURNV( attribute );
	}
 	else
	{
		DalAttribute * attribute;

		attribute = new DalAttribute(
						attributeName,
						units,
						comment,
						this );

		_attributes.push_back( attribute );

		ALTRETURNV( attribute );
	} // if
}


void
DalAttributable::addAttributes( const Attributable * org )
{
	PROC

	//operator=( *org );

	for (unsigned i = 0; i < org -> attributes(); ++i)
		addAttribute( org -> attribute(i) );

	for (unsigned i = 0; i < org -> comments(); ++i)
		addComment( org -> comment(i) );

	for (unsigned i = 0; i < org -> historys(); ++i)
		addHistory( org -> history(i) );


	ENDPROC
}


string
DalAttributable::key(const string& name) const
{
  PROC
  string k = uppercase(name);
  if (k.length() > 8) k = k.substr(0,8);
  RETURN k;
}


bool
DalAttributable::hasAttribute( const string & attributeName ) const
{

	PROC

	const string	attName = DalAttributable::parseName( attributeName );
	const string	k = key(attName);

	for( AttributeList::const_iterator i = _attributes.begin();
		 i != _attributes.end(); ++i )
	{
		if( key( (*i) -> name() ) == k )
		{
			ALTRETURNV( true );
		} // if
	} // for

	RETURN false;

}


const Attribute *
DalAttributable::attribute(
			const string &	attributeName,
			AccessMode	accessMode) const
{

	PROC
	RETURN attribute( attributeNumber( attributeName ), accessMode );
}


Attribute *
DalAttributable::attribute(
			const string&	attributeName,
			AccessMode	accessMode )
{
	PROC
	RETURN attribute( attributeNumber( attributeName ), accessMode );
}


const Attribute *
DalAttributable::attribute(
			unsigned int	number,
			AccessMode	accessMode ) const
{
	PROC

	checkValidAttributeNumber( number );
	DalAttribute * att = _attributes[ number ];
	att -> mode( accessMode );
	att -> alwaysAllowModify( false );

	RETURN const_cast<DalAttribute*>( att );

}


Attribute *
DalAttributable::attribute(
			unsigned int	number,
			AccessMode	accessMode )
{
	PROC

	checkValidAttributeNumber( number );
	DalAttribute * att = _attributes[ number ];
	att -> mode( accessMode );

	RETURN att;

}


const string &
DalAttributable::comment( unsigned int number ) const
{
	PROC
	RETURN _comments[ number ];

}


const string &
DalAttributable::history ( unsigned int number ) const
{
	PROC
	RETURN _history[ number ];

}


void
DalAttributable::deleteAttribute( const string & name )
{
	PROC
	deleteAttribute( attributeNumber( name ) );
	ENDPROC
}


void
DalAttributable::deleteAttribute( unsigned int number )
{
	PROC

	checkIfReadOnly();

	if( number >= _attributes.size() )
	{
		errstr << "It was not possible to delete the attribute with number '" << number << "' as it could not be found. The number of attributes in the set with qualified name " << qualifiedName() << " is actually '" << _attributes.size() << "'" << Sas::error("attributeNotFound");

	}

	AttributeList::iterator i = _attributes.begin();

	for( unsigned int j = 0; j < number; ++j, ++i );

	delete( *i );

	_attributes.erase( i );

	ENDPROC
}


unsigned int
DalAttributable::attributeNumber( const string & attributeName ) const
{
	PROC

	SAS_DEBUG( clog << "looking up attribute "<<attributeName<<" in attributable set " << name() << endl; )

	const string	attName = DalAttributable::parseName( attributeName );
	const string	k = key(attName);

	unsigned int attNumber = 0;

	if( !attName.empty() )
	{
 		AttributeList::const_iterator i;

		for (i = _attributes.begin(); i != _attributes.end(); ++i, ++attNumber)
		{
			SAS_DEBUG( clog << "encountered attribute " << (*i)->name() << endl; )

			if( key( (*i)->name() ) == k )
			{
				break;

			} // if
		} // for

		if( i == _attributes.end() )
		{
			errstr << "Could not find attribute with qualified name " << qualifiedName() << "%" << attName << " in attributable set " << name() << Sas::error("attributableNotFound");

		} // if
	} // if

	RETURN attNumber;
}


void
DalAttributable::printOn( ostream & os ) const
{
	PROC

	// ****************
	// Write Attributes
	// ****************

	os << "There are " << _attributes.size() << " Attributes" << endl;

	unsigned int no = 1;

	for(
		AttributeList::const_iterator i = _attributes.begin();
		i != _attributes.end();
		++i, ++no )
	{
		os << "[" << no << "]" << *(*i) << endl;
	} // for


	// **************
	// Write Comments
	// **************

	os << "There are " << _comments.size() << " Comments" << endl;

	no = 1;

	for(
		CommentList::const_iterator i = _comments.begin();
		i != _comments.end();
		++i, ++no )
	{
		os << "[" << no << "]" << "COMMENT: " << (*i).c_str() << endl;
	} // for


	// *************
	// Write History
	// *************

	os << "There are " << _history.size() << " History" << endl;

	no = 1;

	for(
		HistoryList::const_iterator i = _history.begin();
		i != _history.end();
		++i, ++no )
	{
		os << "[" << no << "]" << "HISTORY: " << (*i).c_str() << endl;

	} // for


	ENDPROC
}


string
DalAttributable::parseName( const string & compoundName ) const
{
	PROC

	SimpleScanner	sc( compoundName, ":%" );
	string 		parsedName;

	try
	{
		parsedName = sc.match( SimpleToken::String ).asString();

		if( sc.nextIsCharacter( ':' ) || sc.nextIsCharacter( '%' ) )
		{
			parsedName = parseName( compoundName );
			SimpleScanner	sc( parsedName, ":%" );
			sc.match( '%' );
			parsedName = "";

			if( !sc.eof() )
			{
				parsedName = sc.match( SimpleToken::String ).asString();

				if( !sc.eof() )
				{
					errstr << "The trailing characters " << sc.remainder() << " could not be parsed" << Sas::error( "invalidTrailingCharacters" );
				} // if
			} // if
		} // if
	}
	catch( SimpleScanner::Expected e )
	{
		errstr << "An error was detected with the name '" << compoundName << "'.  " << e.message() << Sas::error( "invalidCompoundName" );
	}

	RETURN parsedName;

}






