using namespace std;
/*************************************************************** 
* 
* DalAttribute.cc             & XMM Science Analysis System (SAS) 
* 
* Class representing an Attribute. 
* 
* (c) 1997 ESA - Design by Marco Beijersbergen, ESA
*                Written by Mark Thomas, CARA
* 
* $Id: DalAttribute.cc,v 1.2 2004/03/05 14:43:00 eojero Exp $
* 
*****************************************************************/ 

#include "DalAttribute.h"
#include <errstr.h>
#include <Tracer.h>
#include <dal_utilities.h>
#include<NAN.h>
#include<sstream>

class Value
{
	public :

		virtual ~Value() {}

		virtual Value & operator=( int value )
				{ PROC cout << value; RETURN *this; }
		virtual Value & operator=( double value )
				{ PROC cout << value; RETURN *this; }
		virtual Value & operator=( bool value )
				{ PROC cout << value; RETURN *this; }   
		virtual Value & operator=( const string & value )
				{ PROC cout << value; RETURN *this; }
		virtual Value & operator=( const Attribute& ) = 0;

		virtual int asInt() const = 0;
		virtual double asReal() const = 0;
		virtual bool asBool() const = 0;
		virtual string asString() const = 0;
		virtual Attribute::DataType dataType() const = 0;
		virtual unsigned dataSize() const = 0;
		virtual char* data() const = 0;

};



class IntValue : public virtual Value
{
	public :

		Value & operator=( int value )
				{ PROC _value = value; RETURN *this; }
		Value& operator=( const Attribute& attribute ) { PROC _value = attribute.asInt(); RETURN *this; }
		int asInt() const { PROC RETURN _value; }
		double asReal() const { PROC RETURN _value; }
		bool asBool() const { PROC RETURN _value; }
		string asString() const { PROC RETURN toString( _value ); }
		Attribute::DataType dataType() const { PROC RETURN Attribute::Int; }
		unsigned dataSize() const { PROC RETURN sizeof(int); }
		char* data() const { PROC RETURN (char*)&_value; }

	private :

		int	_value;
};


class RealValue : public virtual Value
{
public :
    Value & operator=( double value )
	{ PROC
	      _value = value;
	RETURN *this; }
		virtual Value& operator=( const Attribute& attribute ) { PROC _value = attribute.asReal(); RETURN *this; }
		int asInt() const { PROC RETURN int( _value ); }
		double asReal() const { PROC RETURN _value; }
		bool asBool() const { PROC RETURN bool( _value ); }
		string asString() const { PROC RETURN toString( _value ); }
		Attribute::DataType dataType() const { PROC RETURN Attribute::Real; }
		unsigned dataSize() const { PROC RETURN sizeof(double); }
		char* data() const { PROC RETURN (char*)&_value; }

	private :

		double	_value;
};


class BoolValue : public virtual Value
{
	public :

		Value & operator=( bool value )
				{ PROC _value = value; RETURN *this; }
		virtual Value& operator=( const Attribute& attribute ) { PROC
		_value = attribute.asBool(); RETURN *this; }
		int asInt() const { PROC RETURN _value; }
		double asReal() const { PROC RETURN _value; }
		bool asBool() const { PROC RETURN _value; }
		string asString() const { PROC RETURN (_value ? "T" : "F"); }
		Attribute::DataType dataType() const { PROC RETURN Attribute::Bool; }
		unsigned dataSize() const { PROC RETURN sizeof(bool); }
		char* data() const { PROC RETURN (char*)&_value; }

	private :

		bool	_value;
};


class StringValue : public virtual Value
{
	public :

		Value & operator=( const string & value )
				{ PROC _value = value; RETURN *this; }
		virtual Value& operator=( const Attribute& attribute ) { PROC _value = attribute.asString(); RETURN *this; }
		int asInt() const { PROC RETURN atoi( _value.c_str() ); }
		double asReal() const { PROC RETURN atof( _value.c_str() ); }
		bool asBool() const { PROC RETURN (_value=="T" ? 1 : 0); }
		string asString() const { PROC RETURN _value; }
		Attribute::DataType dataType() const { PROC RETURN Attribute::DString; }
		unsigned dataSize() const { PROC RETURN _value.length(); }
		char* data() const { PROC RETURN (char*)_value.c_str(); }

	private :

		string	_value;
};


DalAttribute::DalAttribute(
		const Attribute*	org,
		const DalAttributable* 	parent )
	 :	DalLabelled( org -> name(), org->label() ),
		DalDal( parent -> mode(), parent ),
		DalChild<Attributable>( parent ),
		_value( 0 ),
		_units( org->units() )
{
	PROC

	SAS_DEBUG( clog << "copying attribute " << *org << endl; )

	switch ( org->dataType() ) {

		case Attribute::Bool:    operator=( org->asBool() );   break;
		case Attribute::Int:     operator=( org->asInt() );    break;
		case Attribute::Real:    operator=( org->asReal() );   break;
		case Attribute::DString: operator=( org->asString() ); break;

		default: errstr << "could not copy attribute " << org->name() <<" from " << org->parent()->name() << " since its data type " << int( org->dataType() ) <<" was not recognised" << Sas::error("invalidAttributeDataType");
	}

	//DEBUG( clog << "copy constructed DalAttribute " << this << ": " << *this << endl; )
	ENDPROC
}


DalAttribute::DalAttribute(
			const string &		name,
			const string &		units,
			const string &		comment,
			const DalAttributable *	parent )

			:	DalLabelled( name, comment ),
				DalDal( parent -> mode(), parent ),
				DalChild<Attributable>( parent ),
				_value( 0 ),
				_units( units )
{
	PROC
	//DEBUG( clog << "constructed DalAttribute " << this << ": " << *this << endl; )
	ENDPROC
}


DalAttribute::~DalAttribute()
{
	PROC
	delete _value;
	ENDPROC
}


void
DalAttribute::dataType(Attribute::DataType type)
{
	PROC

	checkIfReadOnly();
	delete _value;
	
	switch( type )
	{
		case Attribute::Bool :
			_value = new BoolValue;
			break;
		case Attribute::Int :
			_value = new IntValue;
			break;
		case Attribute::Real :
			_value = new RealValue;
			break;
		case Attribute::DString :
			_value = new StringValue;
			break;
		default :
			errstr << "invalid data type " << int( type ) << " specified for the attribute " << name() << " of " << parent()->name() <<Sas::error("invalidAttributeDataType");
				break;
	} // switch

	ENDPROC
}


Attribute &
DalAttribute::operator=( int	value )
{
	PROC

	checkIfReadOnly();
	dataType(Attribute::Int);
	*_value = value;

	RETURN *this;

}


Attribute &
DalAttribute::operator=( double value )
{
	PROC

	checkIfReadOnly();
	dataType(Attribute::Real);

	if(NaN.isSet(value)){
	    errstr << "There was an attempt to set the attribute "
		   << qualifiedName() << " whose value was " << asReal()
		   << " to NaN. This is not possible, therefore the value will be cleared. The old value can be found in the comment string."
		   << Sas::warning("AttributeNaN");
	    stringstream ss;
	    ss << setprecision(13) << "AttributeNaN. Last value " << asReal();
	    relabel(ss.str());
	    *this->parent()->attribute(name()) = "";
	    
	} else
	    *_value = value;

	RETURN *this;

}


Attribute &
DalAttribute::operator=( const string & value )
{
	PROC

	checkIfReadOnly();
	dataType(Attribute::DString);
	*_value = value;

	RETURN *this;

}

Attribute &
DalAttribute::operator=( const char * value )
{
	PROC

	operator=(string(value));

	RETURN *this;

}


Attribute &
DalAttribute::operator=( bool value )
{
	PROC

	checkIfReadOnly();
	dataType(Attribute::Bool);
	*_value = value;

	RETURN *this;

}


Attribute &
DalAttribute::operator=( const Attribute & value )
{
	PROC

	relabel( value.label() );
	_units = value.units();
	dataType( value.dataType() );
	*_value = value;

	RETURN *this;
}


void DalAttribute::printOn( ostream & os ) const
{
	PROC

	os << "ATTRIBUTE: name='" << qualifiedName() << "' units='" << _units << "' comment='" << label() << "'";

	if ( _value )
		os << " dataType='" << toString( dataType() ) << "' value='" << asString() << "'";
	else
		os << " (no value)";

	ENDPROC
}


int
DalAttribute::asInt() const
{
	PROC
	checkValue();
	RETURN _value->asInt();
}


double
DalAttribute::asReal() const
{
	PROC
	checkValue();
	RETURN _value->asReal();
}


string
DalAttribute::asString() const
{
	PROC
	checkValue();
	RETURN _value->asString();
}


bool
DalAttribute::asBool() const
{
	PROC
	checkValue();
	RETURN _value->asBool();
}


const string &
DalAttribute::units() const
{
	PROC
	RETURN _units;
}


void
DalAttribute::units(const string& u)
{
	PROC
	checkIfReadOnly();
	_units = u;
	ENDPROC
}


DalAttribute::DataType
DalAttribute::dataType() const
{
	PROC
	checkValue();
	RETURN _value -> dataType();
}


char*
DalAttribute::addressOfValue() const
{
	PROC
	checkValue();
	RETURN _value->data();
}


unsigned int
DalAttribute::dataSize() const
{
	PROC
	checkValue();
	RETURN _value->dataSize();
}


unsigned int
DalAttribute::attributeNumber() const
{
	PROC
	RETURN parent() -> attributeNumber( name() );
}


void
DalAttribute::checkValue() const
{
	PROC
	if( !_value) errstr << "The attribute with name " << name() << " has not yet been assigned a value, and is thus not accessible" << Sas::error( "NoValueAttribute" );
	ENDPROC
}


string
DalAttribute::parseName( const string & compoundName ) const
{
	PROC
	RETURN compoundName;
}


string
DalAttribute::qualifiedName() const
{
	PROC
	RETURN parent() -> qualifiedName() + "%" + name();
}


void
DalAttribute::release()
{
	PROC
	ENDPROC
}


