using namespace std;
/*************************************************************** 
* 
* HiLowArray.cc                & XMM Science Analysis System (SAS) 
* 
* Class representing an Array.
* 
* (c) 1997 ESA - Written by Mark Thomas, CARA
* 
* $Id: HiLowArray.cc,v 1.2 2004/03/05 14:43:02 eojero Exp $
* 
*****************************************************************/ 


#include "HiLowArray.h"
#include "HiLowArrayData.h"
#include "Storage.h"

#include <errstr.h>
#include <Tracer.h>

typedef Storage<HiLowArrayData> ArrayDataStorage;


static unsigned long
numberOfElements( const longVec &	size )
{
	PROC
	unsigned long elements = 1;
	for( unsigned int i = 0; i < size.size(); ++i ) elements *= size[ i ];
	RETURN elements;
}


HiLowArray::HiLowArray(
		const string &			name,
		DataType 			dataType,
		const longVec &			size,
		const string &			units,
		const string &			label,
		DalDataSet *			parent )

	:	DalArray( name, dataType, size, units, label, parent ),
		DalSeekable< longVec >( zero, size ),
		_data( new ArrayDataStorage(
				parent->name() + "." + name,
				dataType,
				::numberOfElements( size ),
				0 ))
{
	PROC
	ENDPROC

}


HiLowArray::HiLowArray(
		const Array &			array,
		DalDataSet *			parent,
		string				newName)
	:	DalArray(
			newName.empty() ? array.name() : newName,
			array.dataType(),
			array.size(),
			array.units(),
			array.label(),
			parent ),
		DalSeekable< longVec >( zero, array.size() )
{
	PROC

	operator=( array );

	ENDPROC

}


HiLowArray::~HiLowArray()
{
	PROC

	// *********************************
	// Release and delete thearray data. 
	// *********************************
	//release();

	checkUnreleasedData();

	delete _data;

	ENDPROC
}


HiLowArray &
HiLowArray::operator=( const Array & array )
{
	PROC

	//relabel(array.label());

	(Attributable&)*this = array;

	delete _data;
	_data = new ArrayDataStorage(
			name(),
			array.dataType(),
			::numberOfElements( size() ),
			0 );

	ArrayData * s = array.data();
	ArrayData * d = data();
	unsigned long N = elements();

	switch( array.dataType() )
	{
		case Bool :
		{
			bool8 * sd = s -> boolData();
			bool8 * dd = d -> boolData();
			for (unsigned long i = 0; i < N; ++i)
				dd[i] = sd[i];
		}
			break;
		case Int8 :
		{
			int8 *sd = s->int8Data();
			int8 *dd = d->int8Data();
			for (unsigned long i = 0; i < N; ++i)
				dd[i] = sd[i];
		}
			break;
		case Uint16 :
		{
			uint16 *sd = s->uint16Data();
			uint16 *dd = d->uint16Data();
			for (unsigned long i = 0; i < N; ++i)
				dd[i] = sd[i];
		}
			break;
		case Int16 :
		{
			int16 *sd = s->int16Data();
			int16 *dd = d->int16Data();
			for (unsigned long i = 0; i < N; ++i)
				dd[i] = sd[i];
		}
			break;
		case Uint32 :
		{
			uint32 *sd = s->uint32Data();
			uint32 *dd = d->uint32Data();
			for (unsigned long i = 0; i < N; ++i)
				dd[i] = sd[i];
		}
			break;
		case Int32 :
		{
			int32 *sd = s->int32Data();
			int32 *dd = d->int32Data();
			for (unsigned long i = 0; i < N; ++i)
				dd[i] = sd[i];
		}
			break;
		case Real32 :
		{
			real32 *sd = s->real32Data();
			real32 *dd = d->real32Data();
			for (unsigned long i = 0; i < N; ++i)
				dd[i] = sd[i];
		}
			break;
		case Real64 :
		{
			real64 *sd = s->real64Data();
			real64 *dd = d->real64Data();
			for (unsigned long i = 0; i < N; ++i)
				dd[i] = sd[i];
		}
			break;

		default:
			errstr << "invalid type " << int( array.dataType() ) << " in array " << array.name() << Sas::error("internalError");
	} // switch

	delete s;
	delete d;

	RETURN *this;
}


unsigned long
HiLowArray::elements() const
{
	PROC
	RETURN numberOfElements( size() );

}


unsigned int
HiLowArray::dataSize() const
{
	PROC
	RETURN _data -> dataSize();
}


ArrayData *
HiLowArray::makeArrayData( AccessMode			accessMode ) const
{
	PROC
	RETURN new HiLowArrayData(
				_data,
				this,
				DalSeekable< longVec >::from(),
				DalSeekable< longVec >::count(),
				accessMode );

}


/*
ArrayData *
HiLowArray::data(
		const longVec &			from = zero,
		const longVec &			count = zero,
		AccessMode			accessMode=AsParent ) const
{
	PROC
	RETURN new HiLowArrayData(
				_data,
				this,
				DalSeekable< longVec >::from(),
				DalSeekable< longVec >::count(),
				accessMode );

}
*/

#ifdef OBSOLETE

void
HiLowArray::scaleData( 
		double			zero,
		double			scale,
		DataType		toType )
{

	PROC

	HiLowArray		newArray(
					name(),
					toType,
					size(),
					parent() );

	// ************************************
	// Set the following:
	//   scaling
	//   data
	// ************************************
	newArray.set( zero, scale, toType );
	newArray.setData( *this );

	*this = newArray;

	ENDPROC

}


void
HiLowArray::unscaleData( DataType toType )
{

	PROC

	HiLowArray		newArray(
					name(),
					toType,
					size(),
					parent() );

	// ************************************
	// Set the following:
	//   scaling
	//   data
	// Unset
	//   scale
	// ************************************
	newArray.set( -zero() / scale(), 1.0 / scale(), toType );
	newArray.setData( *this );
	newArray.unset();

	*this = newArray;

	ENDPROC

}


template<class T, class S> static void
copyArray(
	T *			to,
	S *			from,
	unsigned long		noOfElementsToCopy,
	const DataComponent &	scale )
{
	PROC

	for( unsigned long i = 0; i < noOfElementsToCopy; ++i )
	{
		to[ i ] = (T)((from[ i ] - scale.zero())/ scale.scale());

	} // for

	ENDPROC

}


template< class T> static void
copyData(
	HiLowArray &			toThisArray,
	const T *			addressOfDataToCopy )
{
	PROC

	switch( toThisArray.dataType() )
	{
		case Array::Int8 :

			copyArray(
				toThisArray.data() -> int8Data(), 
				addressOfDataToCopy,
				toThisArray.elements(),
				toThisArray );

			break;

		case Array::Int16 :

			copyArray(
				toThisArray.data() -> int16Data(), 
				addressOfDataToCopy,
				toThisArray.elements(),
				toThisArray );

			break;

		case Array::Int32 :

			copyArray(
				toThisArray.data() -> int32Data(), 
				addressOfDataToCopy,
				toThisArray.elements(),
				toThisArray );

			break;

		case Array::Real32 :

			copyArray(
				toThisArray.data() -> real32Data(), 
				addressOfDataToCopy,
				toThisArray.elements(),
				toThisArray );

			break;

		case Array::Real64 :

			copyArray(
				toThisArray.data() -> real64Data(), 
				addressOfDataToCopy,
				toThisArray.elements(),
				toThisArray );

			break;

		default :

			errstr << "array " << toThisArray.name() << " has an invalid data type " << toThisArray.dataType() << Sas::error("invalidDataType");

			break;

	} // switch

	ENDPROC

}


void
HiLowArray::setData( const HiLowArray &	usingDataInThisArray )
{
	PROC

	switch( usingDataInThisArray.dataType () )
	{
		case Array::Int8 :

			copyData(
				*this,
				usingDataInThisArray.data() -> int8Data() );
			break;

		case Array::Int16 :

			copyData(
				*this,
				usingDataInThisArray.data() -> int16Data() );

			break;

		case Array::Int32 :

			copyData(
				*this,
				usingDataInThisArray.data() -> int32Data() );

			break;

		case Array::Real32 :

			copyData(
				*this,
				usingDataInThisArray.data() -> real32Data() );

			break;

		case Array::Real64 :

			copyData(
				*this,
				usingDataInThisArray.data() -> real64Data() );

			break;

		default :

			errstr << "Array " << usingDataInThisArray.name() << " has an invalid data type " << usingDataInThisArray.dataType() << Sas::error("invalidDataType");

			break;

	} // switch

	ENDPROC

}

#endif

void
HiLowArray::printOn( ostream & os ) const
{
	PROC

	// ****************
	// Write the Block.
	// ****************
	DalBlock::printOn( os );

	// ***************************
	// Write the Array Dimensions.
	// ***************************
	os << "There are " << dimensions() << " dimensions " << endl;
	os << "Dimensions:";

	for( unsigned int i = 0; i < dimensions(); ++i )
	{
		os << " [" << i + 1 << "]->" << size()[ i ];
	} // for

	os << endl;

	// *********************
	// Write the Array Data.
	// *********************

	ENDPROC
}


//void
//HiLowArray::release()
//{
//	PROC
	//_data -> releaseAll();
//	while( _dataObjects.size() )
//	{
//		errstr << "Unreleased data detected for array with qualified name " << qualifiedName() << Sas::warning( "unreleasedData" );
//		delete _dataObjects[ _dataObjects.size() - 1 ];
//	} // while
//	ENDPROC
//}


