using namespace std;
/*************************************************************** 
 * 
 * MemColumn.cc               & XMM Science Analysis System (SAS) 
 * 
 * Class representing Table Column.
 * 
 * (c) 1997 ESA - Written by Mark Thomas, CARA
 * 
 * $Id: MemColumn.cc,v 1.2 2004/03/05 14:43:03 eojero Exp $ 
 * 
 *****************************************************************/ 


#include <errstr.h>
#include <Tracer.h>
#include "machine_specific_types.h"

#include "MemColumn.h"
#include "MemColumnData.h"
#include "MemCellData.h"


static unsigned long
numberOfElements( const vector<unsigned long> & size )
{
    PROC

	unsigned long elements = 0;

    if( !( size == Dal::zero ) )
    {
	elements = 1;
	for( unsigned int i = 0; i < size.size(); ++i )
	    elements *= size[ i ];
    } // if

    RETURN elements;
}



MemColumn::MemColumn(
    const string &			name,
    DataType			dataType,
    const string &			label,
    const string &			units,
    const vector<unsigned long> &	size,
    DalTable *			parent )

    :
    DalColumn( name, dataType, size, label, units, parent ),
    DalSeekable<unsigned long>(
	parent -> from(),
	parent -> count() )
{

    PROC

	errstr << "Constructing MemColumn with qualified name='"
	       << qualifiedName() <<"' dataType="<<int( dataType )
	       << " rows="<<parent->rows()<<" dimensions="
	       << dimensions() <<" units="<<units
	       << Sas::message( Msg::Core, Msg::Verbose );

    if( cellType() == Fixed )
    {
	_data = new (DataStorage *)[ 1 ];
	_data[ 0 ] = new DataStorage(
	    parent->name()+"."+name,
	    dataType,
	    parent -> rows() * ::numberOfElements(
		size ) );
    }
    else
    {
	_data = new (DataStorage *)[ parent -> rows() ];
	for( unsigned long i = 0; i < parent -> rows(); ++i )
	{
	    _data[ i ] = new DataStorage(
		parent->name()+"."+name,
		dataType,
		0 );

	} // for
    } // if

    ENDPROC

	}


MemColumn::~MemColumn()
{
    PROC

	checkUnreleasedData();

    if( cellType() == Fixed )
    {
	delete _data[ 0 ];
    }
    else
    {
	for( unsigned int i = 0; i < DalChild<Table>::parent() -> rows(); ++i )
	{
	    delete _data[ i ];
	} // for
    } // if
    delete[] _data;
    ENDPROC
	}


MemColumn::MemColumn(
    const DalColumn &	column,
    DalTable * 		parent,
    string 			newName)
    : 	
    DalColumn( column, parent, newName ),
    DalSeekable<unsigned long>( parent -> from(), parent -> count() )
{
    PROC

	// ****************************************************
	// The following allocations are done, so that
	// column=() can be called to do the actual copying.
	// In fact, column=() deletes this memory
	// which is really a waste of time.
	// There is probably a more efficient way of doing this
	// e.g. in column=(), could do the check
	//		if( _data ) delete _data[ 0 ];
	// Then it would only be necessary to set _data = 0
	// in this constructor before calling column=().
	// ****************************************************
	if( column.cellType() == Fixed )
	{
	    _data = new (DataStorage *)[ 1 ];
	    _data[ 0 ] = new DataStorage(
		name()+"."+newName,
		column.dataType(),
		parent -> rows() * ::numberOfElements( column.size()) );
	}
	else
	{
	    _data = new (DataStorage *)[ parent -> rows() ];
	    for( unsigned long i = 0; i < parent -> rows(); ++i )
	    {
		_data[ i ] = new DataStorage(
		    name()+"."+newName,
		    dataType(),
		    0 );

	    } // for
	} // if

    operator=(column);

    ENDPROC;

}


void
MemColumn::assignFixed( const Column & column )
{
    PROC

	delete _data[ 0 ];
    _data[ 0 ] = new DataStorage(
	name(),
	column.dataType(),
	rows() * ::numberOfElements( size() ) );

    ColumnData *s = column.data();
    ColumnData *d = data();
    unsigned long N = rows() * elements();

    switch( dataType() )
    {
	case Bool :
	{
	    bool8 *sd = s->boolData();
	    bool8 *dd = d->boolData();
	    for (unsigned long i = 0; i < N; ++i)
		dd[i] = sd[i];
	}
	break;

	case Int8:
	{
	    int8 *sd = s->int8Data();
	    int8 *dd = d->int8Data();
	    for (unsigned long i = 0; i < N; ++i)
		dd[i] = sd[i];
	}
	break;

	case Uint16:
	{
	    uint16 *sd = s->uint16Data();
	    uint16 *dd = d->uint16Data();
	    for (unsigned long i = 0; i < N; ++i)
		dd[i] = sd[i];
	}
	break;

	case Int16:
	{
	    int16 *sd = s->int16Data();
	    int16 *dd = d->int16Data();
	    for (unsigned long i = 0; i < N; ++i)
		dd[i] = sd[i];
	}
	break;

	case Uint32:
	{
	    uint32 *sd = s->uint32Data();
	    uint32 *dd = d->uint32Data();
	    for (unsigned long i = 0; i < N; ++i)
		dd[i] = sd[i];
	}
	break;

	case Int32:
	{
	    int32 *sd = s->int32Data();
	    int32 *dd = d->int32Data();
	    for (unsigned long i = 0; i < N; ++i)
		dd[i] = sd[i];
	}
	break;

	case Real32:
	{
	    real32 *sd = s->real32Data();
	    real32 *dd = d->real32Data();
	    for (unsigned long i = 0; i < N; ++i)
		dd[i] = sd[i];
	}
	break;
	case Real64:
	{
	    real64 *sd = s->real64Data();
	    real64 *dd = d->real64Data();
	    for (unsigned long i = 0; i < N; ++i)
		dd[i] = sd[i];
	}
	break;

	case DString:
	{
	    char *sd = s->stringData();
	    char *dd = d->stringData();
	    for (unsigned long i = 0; i < N; ++i)
		dd[i] = sd[i];
	}
	break;

	default:
	    errstr << "invalid type " << int( dataType() ) << " in column " << column.name() << Sas::error("internalError");
    } // switch

    delete d;
    delete s;

    ENDPROC
	}


void
MemColumn::assignVariable( const Column & column )
{
    PROC
	for(unsigned k = 0; k < column.rows(); ++k)
	    delete _data[k];
  
    _data = new (DataStorage *)[column.rows()];
    for( unsigned long i = 0; i < column.rows(); ++i ){
	CellData * s = column.cellData(i);
	_data[i] = new DataStorage(name(),
				   column.dataType(),
				   s->size());
	CellData * d = cellData(i);
	switch(column.dataType()) {
	    case Bool:
	    {
		bool8 * sd = s->boolData();
		bool8 * dd = d->boolData();
		for(unsigned long k = 0; k < s->size(); ++k)
		    dd[k] = sd[k];
		break;
	    }
	    case Int8:
	    {
		int8 * sd = s->int8Data();
		int8 * dd = d->int8Data();
		for(unsigned long k = 0; k < s->size(); ++k)
		    dd[k] = sd[k];
		break;
	    }
	    case Int16:
	    {
		int16 * sd = s->int16Data();
		int16 * dd = d->int16Data();
		for(unsigned long k = 0; k < s->size(); ++k)
		    dd[k] = sd[k];
		break;
	    }
	    case Uint16:
	    {
		uint16 * sd = s->uint16Data();
		uint16 * dd = d->uint16Data();
		for(unsigned long k = 0; k < s->size(); ++k)
		    dd[k] = sd[k];
		break;
	    }
	    case Int32:
	    {
		int32 * sd = s->int32Data();
		int32 * dd = d->int32Data();
		for(unsigned long k = 0; k < s->size(); ++k)
		    dd[k] = sd[k];
		break;
	    }
	    case Uint32:
	    {
		uint32 * sd = s->uint32Data();
		uint32 * dd = d->uint32Data();
		for(unsigned long k = 0; k < s->size(); ++k)
		    dd[k] = sd[k];
		break;
	    }
	    case Real32:
	    {
		real32 * sd = s->real32Data();
		real32 * dd = d->real32Data();
		for(unsigned long k = 0; k < s->size(); ++k)
		    dd[k] = sd[k];
		break;
	    }
	    case Real64:
	    {
		real64 * sd = s->real64Data();
		real64 * dd = d->real64Data();
		for(unsigned long k = 0; k < s->size(); ++k)
		    dd[k] = sd[k];
		break;
	    }
	    case DString:
	    {
		char * sd = s->stringData();
		char * dd = d->stringData();
		for(unsigned long k = 0; k < s->size(); ++k)
		    dd[k] = sd[k];
		break;
	    }
	}

	delete d;
	delete s;
    }
			       
    ENDPROC
	}


DalColumn &
MemColumn::operator=( const Column & column )
{

    PROC

	(Attributable&)*this = column;
    //(DalColumn&)*this = column;
    //relabel(column.label());

    //_cellType = column.cellType();
    //_dataType = column.dataType();
    //_size = column.size();

    if( cellType() == Fixed )
    {
	assignFixed( column );
    }
    else
    {
	assignVariable( column );
    } // if

    RETURN *this;

}


ColumnData *
MemColumn::makeColumnData(
    unsigned long			from,
    unsigned long			count,
    AccessMode			accessMode ) const
{
    PROC
	RETURN new MemColumnData(
	    _data[ 0 ],
	    this,
	    from,
	    count,
	    accessMode );

}


CellData *
MemColumn::makeCellData(
    unsigned long			rowNumber,
    unsigned long			from,
    unsigned long			count,
    AccessMode			accessMode ) const
{
    PROC
	RETURN new MemCellData(
	    _data[ rowNumber ],
	    this,
	    rowNumber,
	    from,
	    count,
	    accessMode );

}


/*
  ColumnData *
  MemColumn::data(
  unsigned long			from=0,
  unsigned long			count=0,
  AccessMode			accessMode=AsParent ) const
  {
  PROC

  // *********************************************
  // Check that the column is not variable length.
  // *********************************************
  if( cellType() == Variable )
  {
  errstr << "It is not possible to call the data() function on a variable length column: " << name() << Sas::error( "invalidColumnOperation" );
  } // if

  if ( accessMode == AsParent )
  accessMode = this->DalChild<Table>::parent()->parent()->mode();

  RETURN new MemColumnData(
  _data[ 0 ],
  this,
  from,
  count,
  accessMode );
  }


  CellData *
  MemColumn::cellData(
  unsigned long			rowNumber,
  unsigned long			from=0,
  unsigned long			count=0,
  AccessMode			accessMode=AsParent ) const
  {
  PROC

  // ***********************************
  // Check that the row number is valid.
  // ***********************************
  const unsigned long rows = DalChild<Table>::parent() -> rows();

  if( rowNumber >= rows )
  {
  errstr << "The specified row number " << rowNumber << " is invalid for column: " << name() << ", which has " << rows << " rows. The row number must be in the range 0 to " << rows - 1 << "." << Sas::error( "invalidRowNumber " );
  } // if


  // ******************************************
  // Check that the column is not fixed length.
  // ******************************************
  if( cellType() == Fixed )
  {
  errstr << "The cellData() function may not be called on the fixed length column: " << name() << Sas::error( "invalidColumnOperation" );

  } // if

  if ( accessMode == AsParent )
  accessMode = this->DalChild<Table>::parent()->parent()->mode();

  RETURN new MemCellData(
  _data[ rowNumber ],
  this,
  rowNumber,
  from,
  count,
  accessMode );

  }
*/

void
MemColumn::insertRows( unsigned long pos, unsigned long count )
{
    PROC

	checkIfReadOnly();

    DalSeekable<unsigned long>::seek(
	this -> DalChild<Table>::parent() -> from(),
	this -> DalChild<Table>::parent() -> count() );

    if(cellType() == Fixed)
	_data[ 0 ]->insertBytes( pos * elements(), count * elements() );
    else {
    
	// At this stage rows() is already reporting the new number of
	// rows. rows() is implemented via the parent table, and the
	// parent table's row number has been adjusted in DalTable.cc.
	//
	// The rows from 0 to rows() - count - 1 were set before inserting
	// the new rows. The rows from rows() - count to rows() - 1
	// contain nothing, as we haven't yet created them. Further, we
	// need to move the data as follows:
	//  newdata[k] = _data[k] for k = 0 ... pos -1
	//  newdata[k] empty for k = pos ... pos + count - 1
	//  newdata[k] = _data[k-(pos+count)] for k = pos + count ... rows() - 1
    
	DataStorage ** newdata = new (DataStorage *)[rows()];
	for(unsigned long i = 0; i < rows(); ++i){
	    bool dodata = true;
	    unsigned long k = 0;
	    if(i < pos)
		k = i;
	    else if (i >= pos + count)
		k = i - (pos + count);
	    else
		dodata = false;
      
	    unsigned long size = 0;
	    CellData * s = 0;
	    if(dodata){
		s = cellData(k);
		size = s->size();
	    }
	    newdata[i] = new DataStorage(name(),
					 dataType(),
					 size);
	    if(dodata){
		switch(dataType()){
		    case Bool:
		    {
			bool8 * sd = s->boolData();
			bool8 * dd = newdata[i]->boolData(0, size);
			for(unsigned long k = 0; k < s->size(); ++k)
			    dd[k] = sd[k];
			break;
		    }
		    case Int8:
		    {
			int8 * sd = s->int8Data();
			int8 * dd = newdata[i]->int8Data(0, size);
			for(unsigned long k = 0; k < s->size(); ++k)
			    dd[k] = sd[k];
			break;
		    }
		    case Int16:
		    {
			int16 * sd = s->int16Data();
			int16 * dd = newdata[i]->int16Data(0, size);
			for(unsigned long k = 0; k < s->size(); ++k)
			    dd[k] = sd[k];
			break;
		    }
		    case Uint16:
		    {
			uint16 * sd = s->uint16Data();
			uint16 * dd = newdata[i]->uint16Data(0, size);
			for(unsigned long k = 0; k < s->size(); ++k)
			    dd[k] = sd[k];
			break;
		    }
		    case Int32:
		    {
			int32 * sd = s->int32Data();
			int32 * dd = newdata[i]->int32Data(0, size);
			for(unsigned long k = 0; k < s->size(); ++k)
			    dd[k] = sd[k];
			break;
		    }
		    case Uint32:
		    {
			uint32 * sd = s->uint32Data();
			uint32 * dd = newdata[i]->uint32Data(0, size);
			for(unsigned long k = 0; k < s->size(); ++k)
			    dd[k] = sd[k];
			break;
		    }
		    case Real32:
		    {
			real32 * sd = s->real32Data();
			real32 * dd = newdata[i]->real32Data(0, size);
			for(unsigned long k = 0; k < s->size(); ++k)
			    dd[k] = sd[k];
			break;
		    }
		    case Real64:
		    {
			real64 * sd = s->real64Data();
			real64 * dd = newdata[i]->real64Data(0, size);
			for(unsigned long k = 0; k < s->size(); ++k)
			    dd[k] = sd[k];
			break;
		    }
		    // Can FITS support these??
		    case DString:
		    {
			char * sd = s->stringData();
			char * dd = newdata[i]->stringData(0, size);
			for(unsigned long k = 0; k < s->size(); ++k)
			    dd[k] = sd[k];
			break;
		    }
		}
		delete _data[k];
	    }
	}
	_data = newdata;
    }
    ENDPROC
	}

void
MemColumn::deleteRows( unsigned long from, unsigned long count )
{
    PROC

	checkIfReadOnly();

    DalSeekable<unsigned long>::seek(
	this -> DalChild<Table>::parent() -> from(),
	this -> DalChild<Table>::parent() -> count() );

    if(cellType() == Fixed)
	_data[ 0 ]->deleteBytes( from * elements(), count * elements() );
    else {
	//    Sas::error("MemColumn::deleteRows not implemented for variable length vector columns.", "NotImplemented");

	// At this stage rows() is already reporting the new number of
	// rows. rows() is implemented via the parent table, and the
	// parent table's row number has been adjusted in DalTable.cc.
	//
	// _data has rows() + count elements. Of these only those in the
	// ranges [0, from-1] and [from+count, rows() + count -1] shoulr
	// be copied to the new data storage.
	DataStorage ** newdata = new (DataStorage *)[rows()];
	unsigned new_row_idx = 0;
	for(unsigned long k = 0; k < rows() + count; ++k){
	    if(k < from || k >= from + count){
		unsigned long size = _data[k]->size();
		newdata[new_row_idx] = new DataStorage(name(), dataType(),
						       size); 
	
		switch(dataType()){
		    case Bool:
		    {
			bool8 * sd = _data[k]->boolData(0, size);
			bool8 * dd = newdata[new_row_idx]->boolData(0, size);
			for(unsigned long k = 0; k < size; ++k)
			    dd[k] = sd[k];
			break;
		    }
		    case Int8:
		    {
			int8 * sd = _data[k]->int8Data(0, size);
			int8 * dd = newdata[new_row_idx]->int8Data(0, size);
			for(unsigned long k = 0; k < size; ++k)
			    dd[k] = sd[k];
			break;
		    }
		    case Int16:
		    {
			int16 * sd = _data[k]->int16Data(0, size);
			int16 * dd = newdata[new_row_idx]->int16Data(0, size);
			for(unsigned long k = 0; k < size; ++k)
			    dd[k] = sd[k];
			break;
		    }
		    case Uint16:
		    {
			uint16 * sd = _data[k]->uint16Data(0, size);
			uint16 * dd = newdata[new_row_idx]->uint16Data(0, size);
			for(unsigned long k = 0; k < size; ++k)
			    dd[k] = sd[k];
			break;
		    }
		    case Int32:
		    {
			int32 * sd = _data[k]->int32Data(0, size);
			int32 * dd = newdata[new_row_idx]->int32Data(0, size);
			for(unsigned long k = 0; k < size; ++k)
			    dd[k] = sd[k];
			break;
		    }
		    case Uint32:
		    {
			uint32 * sd = _data[k]->uint32Data(0, size);
			uint32 * dd = newdata[new_row_idx]->uint32Data(0, size);
			for(unsigned long k = 0; k < size; ++k)
			    dd[k] = sd[k];
			break;
		    }
		    case Real32:
		    {
			real32 * sd = _data[k]->real32Data(0, size);
			real32 * dd = newdata[new_row_idx]->real32Data(0, size);
			for(unsigned long k = 0; k < size; ++k)
			    dd[k] = sd[k];
			break;
		    }
		    case Real64:
		    {
			real64 * sd = _data[k]->real64Data(0, size);
			real64 * dd = newdata[new_row_idx]->real64Data(0, size);
			for(unsigned long k = 0; k < size; ++k)
			    dd[k] = sd[k];
			break;
		    }
		    case DString:
		    {
			char * sd = _data[k]->stringData(0, size);
			char * dd = newdata[new_row_idx]->stringData(0, size);
			for(unsigned long k = 0; k < size; ++k)
			    dd[k] = sd[k];
			break;
		    }
		}
		++new_row_idx;
	    }
	    delete _data[k];
	}
	_data = newdata;
    }
    ENDPROC
	}

void
MemColumn::copyRows( unsigned long from, unsigned long to, unsigned long count )
{
    PROC
     
	if(from != to){
	    
	    checkIfReadOnly();
	    
	    if(cellType() == Fixed)
		_data[ 0 ]->copyBytes(
		    from * elements(),
		    to * elements(),
		    count * elements() );
	    else {
		// The target row starts at to and it is increased
		// every time we go around the loop (at the end).
		unsigned target_row_idx = to;
		for(unsigned long source_row_idx = from;
		    source_row_idx < from + count;
		    ++source_row_idx){
		    CellData * s = cellData(source_row_idx);
		    unsigned long size = s->size();
		    delete _data[target_row_idx];
		    _data[target_row_idx] = new DataStorage(name(), dataType(),
							    size);
		    CellData * d = cellData(target_row_idx);
		    
		    switch(dataType()) {
			case Bool:
			{
			    bool8 * sd = s->boolData();
			    bool8 * dd = d->boolData();
			    for(unsigned long k = 0; k < size; ++k)
				dd[k] = sd[k];
			    break;
			}
			case Int8:
			{
			    int8 * sd = s->int8Data();
			    int8 * dd = d->int8Data();
			    for(unsigned long k = 0; k < size; ++k)
				dd[k] = sd[k];
			    break;
			}
			case Int16:
			{
			    int16 * sd = s->int16Data();
			    int16 * dd = d->int16Data();
			    for(unsigned long k = 0; k < size; ++k)
				dd[k] = sd[k];
			    break;
			}
			case Uint16:
			{
			    uint16 * sd = s->uint16Data();
			    uint16 * dd = d->uint16Data();
			    for(unsigned long k = 0; k < size; ++k)
				dd[k] = sd[k];
			    break;
			}
			case Int32:
			{
			    int32 * sd = s->int32Data();
			    int32 * dd = d->int32Data();
			    for(unsigned long k = 0; k < size; ++k)
				dd[k] = sd[k];
			    break;
			}
			case Uint32:
			{
			    uint32 * sd = s->uint32Data();
			    uint32 * dd = d->uint32Data();
			    for(unsigned long k = 0; k < size; ++k)
				dd[k] = sd[k];
			    break;
			}
			case Real32:
			{
			    real32 * sd = s->real32Data();
			    real32 * dd = d->real32Data();
			    for(unsigned long k = 0; k < size; ++k)
				dd[k] = sd[k];
			    break;
			}
			case Real64:
			{
			    real64 * sd = s->real64Data();
			    real64 * dd = d->real64Data();
			    for(unsigned long k = 0; k < size; ++k)
				dd[k] = sd[k];
			    break;
			}
			case DString:
			{
			    char * sd = s->stringData();
			    char * dd = d->stringData();
			    for(unsigned long k = 0; k < size; ++k)
				dd[k] = sd[k];
			    break;
			}
		    }
    
		    delete d;
		    delete s;
		    target_row_idx++;
		}
	    }
	}
    ENDPROC
	}


//void
//MemColumn::release()
//{
//	PROC
//	ENDPROC
//}


void
MemColumn::printOn( ostream & os ) const
{
    PROC

	// **********************
	// Write the Column Name.
	// **********************
	os << "COLUMN: Name->" << name() << " dataType=" << int( dataType() ) << " Label=" << label() << " units=" << units() << " dimensions=" << dimensions() << endl;

    // ****************************
    // Write the column attributes.
    // ****************************
    DalAttributable::printOn( os );

    // ****************************
    // TODO: Write the Column Data.
    // ****************************

    ENDPROC

	}


void
MemColumn::rename( const string & newName )
{
    PROC
	if( !DalChild<Table>::parent() -> hasColumn( newName ))
	{
	    DalLabelled::rename( newName );
	}
	else
	{
	    errstr << "It is not possible to rename the column with name '" << name() << "' to '" << newName << "', because a column with name '" << newName << "' already exists in the table with name '" << DalChild<Table>::parent() -> name() << "'" << Sas::error( "invalidColumnName" );
	} // if
    ENDPROC

	}


unsigned long
MemColumn::elements() const
{
    PROC
	RETURN ::numberOfElements( size() );
}


unsigned int
MemColumn::dataSize() const
{
    PROC
	RETURN _data[0]->dataSize();
}


#ifdef OBSOLETE

void
MemColumn::scaleData(
    double			zero,
    double			scale,
    DataType		toType )
{

    PROC

	MemColumn		newColumn(
	    name(),
	    toType,
	    label(),
	    units(),
	    size(),
	    parent() );


    // ************************************
    // Set the following:
    //   scaling
    //   data
    // ************************************
    newColumn.set( zero, scale, toType );
    newColumn.setData( *this );

    *this = newColumn;

    ENDPROC;

}


void
MemColumn::unscaleData( DataType toType )
{
    PROC

	MemColumn		newColumn(
	    name(),
	    toType,
	    label(),
	    units(),
	    size(),
	    parent() );


    // ************************************
    // Set the following:
    //   scaling
    //   data
    // Unset
    //   scale
    // ************************************
    newColumn.set( -zero() / scale(), 1.0 / scale(), toType );
    newColumn.setData( *this );
    newColumn.unset();

    *this = newColumn;

    ENDPROC;

}


template<class T, class S> static void
copyArray(
    T *			to,
    S *			from,
    unsigned long		noOfElementsToCopy,
    DataComponent &		scale )
{
    PROC

	for( unsigned long i = 0; i < noOfElementsToCopy; ++i )
	{
	    to[ i ] = (T)((from[ i ] - scale.zero())/ scale.scale());

	} // for

    ENDPROC;

}


template< class T> static void
copyData(
    MemColumn &			toThisColumn,
    const T *			addressOfDataToCopy )
{
    PROC

	switch( toThisColumn.dataType() )
	{
	    case Column::Int8 :

		copyArray(
		    toThisColumn.data() -> int8Data(), 
		    addressOfDataToCopy,
		    toThisColumn.rows(),
		    toThisColumn );

		break;

	    case Column::Int16 :

		copyArray(
		    toThisColumn.data() -> int16Data(), 
		    addressOfDataToCopy,
		    toThisColumn.rows(),
		    toThisColumn );

		break;

	    case Column::Int32 :

		copyArray(
		    toThisColumn.data() -> int32Data(), 
		    addressOfDataToCopy,
		    toThisColumn.rows(),
		    toThisColumn );

		break;

	    case Column::Real32 :

		copyArray(
		    toThisColumn.data() -> real32Data(), 
		    addressOfDataToCopy,
		    toThisColumn.rows(),
		    toThisColumn );

		break;

	    case Column::Real64 :

		copyArray(
		    toThisColumn.data() -> real64Data(), 
		    addressOfDataToCopy,
		    toThisColumn.rows(),
		    toThisColumn );

		break;

	    default :

		errstr << "Encountered an invalid data type " << toThisColumn.dataType() << " in column " << toThisColumn.name() << " in table " << toThisColumn.parent()->name() << " in data set " << toThisColumn.parent()->parent()->name() << " when trying to copy column data" << Sas::error("invalidColumnDataType");

		break;

	} // switch

    ENDPROC;

}


void
MemColumn::setData( const MemColumn &	usingDataInThisColumn )
{
    PROC

	switch( usingDataInThisColumn.dataType () )
	{
	    case Column::Int8 :
		copyData(
		    *this,
		    usingDataInThisColumn.data() -> int8Data() );
		break;

	    case Column::Int16 :
		copyData(
		    *this,
		    usingDataInThisColumn.data() -> int16Data() );
		break;

	    case Column::Int32 :
		copyData(
		    *this,
		    usingDataInThisColumn.data() -> int32Data() );
		break;

	    case Column::Real32 :
		copyData(
		    *this,
		    usingDataInThisColumn.data() -> real32Data() );
		break;

	    case Column::Real64 :
		copyData(
		    *this,
		    usingDataInThisColumn.data() -> real64Data() );
		break;

	    default :

		errstr << "Encountered an invalid data type " << usingDataInThisColumn.dataType() << " in column " << usingDataInThisColumn.name() << Sas::error("invalidDataType");

		break;

	} // switch

    ENDPROC;

}

#endif
