using namespace std;

#include <fitsio.h>
#include <unistd.h>
#include <sstream>

#include "dal_utilities.h"
#include "HiLowFitsDatabase.h"
#include "errstr.h"
#include "Tracer.h"
#include "stringutil.h"


// ********************
// Global defined here.
// ********************

static const string attributeSeparator = "XMMSEP";
static const string nameOfPrimaryImage = "PRIMARY";
static const string defaultArrayComment = "The name of this array";
static const string defaultTableComment = "The name of this table";
static const string defaultColumnComment = "The name of this column";
static const string arrayUnitsComment = "Units of the image";


// **********************
// Static functions here.
// **********************
static void
deleteCopyrightMessage( fitsfile *	_fptr,
			int *		status )
{
	PROC

	if( *status != 0 )
	{
		ENDPROC
	} // if

	// **********************************************
	// Delete the copyright message.
	// This will be achieved by deleting all comments
	// in the (newly created) primary header unit.
	// Just keep deleting the COMMENT keywords, until
	// there are no more (indicated by KEY_NO_EXIST).
	// **********************************************
	while( *status == 0 ) fits_delete_key( _fptr, "COMMENT*", status );

	if( *status != KEY_NO_EXIST ) fits_check(_fptr);

	*status = 0;

	ENDPROC
}


HiLowFitsDatabase::HiLowFitsDatabase()
	: _fptr( 0 )
{
	PROC
	unlock();
	ENDPROC
}


HiLowFitsDatabase::~HiLowFitsDatabase()
{
	PROC
	// ***************
	// Close the file.
	// ***************
	if( _fptr )
	{
		fitsCloseFile( _fptr, &status );
		fits_check(_fptr);
		_fptr = 0;
	} // if
	ENDPROC
}


void
HiLowFitsDatabase::addEntry( const DataSet & dataSet )
{
	PROC

	// *************************************
	// If locked then do nothing and return.
	// *************************************
	if( locked() )
	{
		ALTRETURN;
	} // if

	const string		setName = dataSet.name();

	dataSetServer -> clobberDataSet( setName );

	//const bool		exists = dataSetServer -> exists( fileName );

	// ****************************************************
	// Check to see if the dataset may be created.
	// If it may, then attempt to clobber it, if necessary.
	// ****************************************************
	//if( exists && !clobber() )
	//{
	//	errstr << "The dataset with name '" << fileName << "' could not be created because it already exists and clobber is not set." << Sas::error( "datasetAlreadyExists" );

	//}
	//else if( exists )
	//{
		// ====================
		// Clobber the dataset.
		// ====================
		//status = unlink( fileName.c_str() );	// UNIX delete file.

		//if( status != 0 )
		//{
		//	errstr << "The dataset with name '" << fileName << "' could not be created because it already exists and clobber is set. but the clobber failed" << Sas::error("clobberFailed");
		//} // if
	//}
	//else if( clobber() )
	//{
		// ===================================
		// Clobber the file , if there is one.
		// ===================================
		//unlink( fileName.c_str() );	// UNIX delete file.
	//} // if

	// ******************************
	// Attempt to create the dataset.
	// ******************************
	fitsCreateFile( &_fptr, setName.c_str(), &status );

	if( status )
	{

		errstr << "The dataset with name '" << setName << "' could not be created." << Sas::error("createDatasetFailed");

	} // if

	fits_check( _fptr );


	// **********************************************************
	// Insert dummy primary header, which is necessary to prevent
	// an error from occurring when the file is opened again.
	// And then remove fits copyright messages, before closing
	// and then checking that it closed.
	// **********************************************************
	fits_create_img( _fptr, 8, 0, 0, &status );
	deleteCopyrightMessage( _fptr, &status );

	ENDPROC
}


void
HiLowFitsDatabase::addEntry( const Block & block )
{
	PROC

	if( locked() )
	{
		ALTRETURN;
	} // if

	switch( block.type() )
	{
		case Block::TableT :
		{
			const Table * table = block.parent()->table( block.name() );
			addEntry( *table );
		}
			break;
		case Block::ArrayT :
		{
			const Array * array = block.parent()->array( block.name() );
			addEntry( *array );
		}
			break;
		default :
			errstr << "the block '" << block.name() << "' has an invalid type " << int( block.type() ) << Sas::error("internalError");
			break;
	} // switch

	ENDPROC
}


void
HiLowFitsDatabase::addEntry( const Table & table )
{
	PROC

	// *************************************
	// If locked then do nothing and return.
	// *************************************
	if( locked() )
	{
		ALTRETURN;
	} // if

	moveToExtension( table, Before );

	// *************
	// Create table.
	// *************
	// SOC-SPR-2390: this fails if the previous table
// 	fits_insert_btbl(
// 			_fptr,
// 			//BINARY_TBL,
// 			table.rows(),
// 			0,
// 			(char **)0,
// 			(char **)0,
// 			(char **)0,
// 			"",
// 			0,
// 			&status );
	fits_create_tbl(
			_fptr,
			BINARY_TBL,
			table.rows(),
			0,
			(char **)0,
			(char **)0,
			(char **)0,
			"",
			&status );
	fits_check(_fptr);
 
	ENDPROC
}


void
HiLowFitsDatabase::addEntry( const Array & array )
{
	PROC

	// *************************************
	// If locked then do nothing and return.
	// *************************************
	if( locked() )
	{
		ALTRETURN;
	} // if

	// **********************************************************
	// Revert the indices since FITS uses the Fortran convention,
	// i.e. the first index runs fastest.
	// **********************************************************
	unsigned dimensions = array.dimensions();
	long size[dimensions];
	for (unsigned i = 0; i < dimensions; ++i)
		size[i] = array.size()[dimensions-1-i];


	moveToExtension( array, Before );

	// *********************************************
	// Check to see if this is the "PRIMARY" image.
	// This has essentially already been done in the
	// call to blockNumberToHdunum().
	// *********************************************
	const int		hdunum = blockNumberToHdunum( array );

	if( hdunum == 1 )
	{
		// ********************************************
		// Create primary Image; since a dummy header
		// already exists, this is achieved by resizing
		// (note that Fits disallows the deletion of
		// the primary HDU).
		// ********************************************
		fits_resize_img(
			_fptr,
			toFITSBitpix( array.dataType() ),
			array.dimensions(),
			size,
			&status );
		fits_check(_fptr);
	}
	else
	{
		// *************
		// Create Image.
		// *************
		fits_insert_img(
			_fptr,
			toFITSBitpix( array.dataType() ),
			array.dimensions(),
			size,
			&status );
		fits_check(_fptr);

	}// if

	ENDPROC
}


void
HiLowFitsDatabase::addEntry( const Column & column )
{
	PROC
	// *************************************
	// If locked then do nothing and return.
	// *************************************
	if( locked() )
	{
		ALTRETURN;
	} // if

	moveToExtension( *column.parent() );

	stringstream	ssformat;
	const long	repeat = column.elements();

	// ***********************************************
	// Determine format type. This is dependent on
	// the column data type.
	// ***********************************************

	// ========================================
	// Check for variable length column; if not
	// then check for vector column.
	// ========================================
	if( repeat == 0 )
	{
		ssformat << "1P";
	}
	else if (repeat > 1)
	{
		ssformat << repeat;
	} // if

	switch( column.dataType() )
	{
		case Column::Int8 :

			ssformat << "B";

			break;

		case Column::Uint16 :

			ssformat << "U" ;

			break;

		case Column::Int16 :

			ssformat << "I" ;

			break;

		case Column::Uint32 :

			ssformat << "V" ;

			break;

		case Column::Int32 :

			ssformat << "J" ;

			break;

		case Column::Real32 :

			ssformat << "E" ;

			break;

		case Column::Real64 :

			ssformat << "D" ;

			break;

		case Column::Bool :

			ssformat << "L" ;

			break;

		case Column::DString :

			ssformat << "A";
			// ****************************************
			// Work around for bug with variable length
			// ASCII columns; need to give a (default)
			// size to ascii variable length columns
			// so that the default (0) is not 
			// applied. The bug is that once a variable
			// length column gets a size of zero,
			// it is not possible to resize it. In fact
			// a floating exception is raised by fits.
			// ****************************************
			if( repeat == 0 )
			{
				ssformat << "(1)";
			} // if

			break;

		default :

			errstr << "could not write column '" << column.name() << "' of table '" << column.parent()->name() << "' since it has an invalid data type " << int( column.dataType() ) << Sas::error("internalError");

			break;

	} // switch

	// *******************************
	// Create column and check status.
	// *******************************
	const string		columnName = column.name();
	const unsigned int	columnNumber = column.columnNumber();
	string format=ssformat.str();

	fits_insert_col(
			_fptr,
			columnNumber + 1,
			(char*)columnName.c_str(),
			(char*)format.c_str(),
			&status );

	fits_check(_fptr);

	ENDPROC
}


static void
resetMaxRowElements(
	fitsfile *	_fptr,
	unsigned int	columnNumber,
	unsigned long	size )
{
	PROC

	char			tform[ 20 ];

	fits_read_key(
			_fptr,
			TSTRING,
			(char *)( "TFORM" + toString( columnNumber + 1 )).c_str(),
			&tform,
			0,
			&status );
	fits_check(_fptr);


	// **********************************
	// Remove the parenthesis from tform.
	// **********************************
	unsigned int i = 0;
	while( i < strlen( tform ) && tform[ i ] != '(' ) ++i;
	unsigned long cellSize = 0;

	if( i < strlen( tform ) )
	{
		sscanf( tform + i + 1, "%lu", &cellSize );
	} // if

	if( cellSize < size )
	{
		sprintf( tform + i, "(%lu)", size );
	} // if
	//tform[ i ] = '\0';


	// ******************************************
	// Update the TFORM to the new (reset) value.
	// ******************************************
	fits_update_key_str(
			_fptr,
			//TSTRING,
			(char *)( "TFORM" + toString( columnNumber + 1 )).c_str(),
			tform,
			0,
			&status );

	fits_check(_fptr);

	ENDPROC
}


void
HiLowFitsDatabase::setVariableColumnSize( const CellData & cellData )
{
	PROC
	// *************************************
	// If locked then do nothing and return.
	// *************************************
	if( locked() )
	{
		ALTRETURN;
	} // if

	const Column *		column = cellData.parent();
	const unsigned long	rowNumber = cellData.row();
	const string		columnName = column -> name();
	const unsigned int	columnNumber = column -> columnNumber();

	moveToExtension( *cellData.parent() -> parent() );

	// ******************************************
	// Set the max number of elements in the row.
	// ******************************************
	resetMaxRowElements( _fptr, columnNumber, cellData.size() );


	// ***********************************
	// Set the size by writing dummy data.
	// ***********************************
	switch( column -> dataType() )
	{
		case Column::Int8 :
		{
			int8 data[ cellData.size() ];

			fits_write_col_byt(
				_fptr,
				columnNumber + 1,
				rowNumber + 1,
				1, // Needed for Binary Tables
				cellData.size(),
				(unsigned char*)data,
				&status );
		}
			break;

		case Column::Uint16 :
		{
			uint16 data[ cellData.size() ];
			fits_write_col_usht(
				_fptr,
				columnNumber + 1,
				rowNumber + 1,
				1, // Needed for Binary Tables
				cellData.size(),
				data,
				&status );
		}
			break;

		case Column::Int16 :
		{
			int16 data[ cellData.size() ];
			fits_write_col_sht(
				_fptr,
				columnNumber + 1,
				rowNumber + 1,
				1, // Needed for Binary Tables
				cellData.size(),
				data,
				&status );
		}
			break;

		case Column::Uint32 :
		{
			uint32 data[ cellData.size() ];
			// ==========================
			// Use TLONG to be consistant
			// with reading TLONGs.
			// ==========================
			fits_write_col_uint(
				_fptr,
				columnNumber + 1,
				rowNumber + 1,
				1, // Needed for Binary Tables
				cellData.size(),
				data,
				&status );
		}
						
			break;

		case Column::Int32 :
		{
			int32 data[ cellData.size() ];
			// ==========================
			// Use TLONG to be consistant
			// with reading TLONGs.
			// ==========================
			fits_write_col_int(
				_fptr,
				columnNumber + 1,
				rowNumber + 1,
				1, // Needed for Binary Tables
				cellData.size(),
				data,
				&status );
		}
						
			break;

		case Column::Real32 :
		{
			real32 data[ cellData.size() ];
			fits_write_col_flt(
				_fptr,
				columnNumber + 1,
				rowNumber + 1,
				1, // Needed for Binary Tables
				cellData.size(),
				data,
				&status );
		}
						
			break;

		case Column::Real64 :
		{
			real64 data[ cellData.size() ];
			fits_write_col_dbl(
				_fptr,
				columnNumber + 1,
				rowNumber + 1,
				1, // Needed for Binary Tables
				cellData.size(),
				data,
				&status );
		}
						
			break;

		case Column::Bool :
		{
			bool8 data[ cellData.size() ];
			fits_write_col_log(
				_fptr,
				columnNumber + 1,
				rowNumber + 1,
				1, // Needed for Binary Tables
				cellData.size(),
				(char *)data,
				&status );
		}

			break;

		case Column::DString :
		{
			// ******************************************
			// Ensure that the string written to the
			// fits file is null terminated. This
			// is necessary in the variable column length
			// case because cfitsio ignores the nelements
			// parameter.
			// ******************************************

			// *********************************************
			// Create a string which has the correct length.
			// *********************************************
			char	* chars = new char[ cellData.size() + 1 ];
			for(
				unsigned int i = 0; 
				i < cellData.size(); 
				++i ) chars[ i ] = 'A';
			chars[ cellData.size() ] = '\0';

			fits_write_col_str(
				_fptr,
				columnNumber + 1,
				rowNumber + 1,
				1, // Needed for Binary Tables
				1,
				&chars,
				&status );

			delete[] chars;

		}
			break;

		default :

			cerr << "could not write variable length column '" << column->name() << "' of table '" << column->parent()->name() << "' to file '" << fits_file_name( _fptr ) << "' since the column type " << int( column->dataType() ) << " is invalid" << Sas::error("internalError");

			break;

	} // switch

	// Check whether the write succeeded.
	fits_check(_fptr);

	ENDPROC
}


void
HiLowFitsDatabase::deleteEntry( const DataSet & dataSet )
{
	PROC
	if( locked() )
	{
		ALTRETURN;
	} // if

	// ***********************************************
	// The file probably already exists, so delete it.
	// ***********************************************
	const char* fileName = dataSet.name().c_str();	// UNIX delete file.

	status = unlink( fileName );	// UNIX delete file.

	if( status != 0 )
	{
		errstr << "Could not delete dataset with name '" << fileName << "'" << Sas::error("deletingFile");
	} // if

	ENDPROC
}


void
HiLowFitsDatabase::deleteEntry( const Block & block )
{
	PROC
	if( locked() )
	{
		ALTRETURN;
	} // if

	switch( block.type() )
	{
		case Block::TableT :
		{
			const Table * table = block.parent()->table( block.name() );
			deleteEntry( *table );
		}
			break;
		case Block::ArrayT :
		{
			const Array *array = block.parent()->array( block.name() );
			deleteEntry( *array );
		}
			break;
		default :
			errstr << "the block '" << block.name() << "' has an invalid type " << int( block.type() ) << Sas::error("internalError");
			break;
	} // switch

	ENDPROC
}


void
HiLowFitsDatabase::deleteEntry( const Table & table )
{
	PROC
	if( locked() )
	{
		ALTRETURN;
	} // if
	moveToExtension( table );

	fits_delete_hdu( _fptr, 0, &status );
	fits_check(_fptr);

	ENDPROC
}


void
HiLowFitsDatabase::deleteEntry( const Array & array )
{
	PROC
	if( locked() )
	{
		ALTRETURN;
	} // if
	moveToExtension( array );

	const int		hdunum = blockNumberToHdunum( array );

	if( hdunum == 1 )
	{
		//errstr << "Deletion of the primary array with qualified name " << array.qualifiedName() << " is not yet implemented" << Sas::error( "notImplemented" );
		// ********************************************
		// Since FITS does not allow the primary
		// header to be deleted, then we will
		// resize it to zero.
		// ********************************************
		fits_resize_img(
			_fptr,
			toFITSBitpix( array.dataType() ),
			0,
			0,
			&status );
	}
	else
	{
		fits_delete_hdu( _fptr, 0, &status );
	} // if
	fits_check(_fptr);
	ENDPROC
}


void
HiLowFitsDatabase::deleteEntry( const Column & column )
{
	PROC
	// *************************************
	// If locked then do nothing and return.
	// *************************************
	if( locked() )
	{
		ALTRETURN;
	} // if
	moveToExtension( *column.parent() );

	// ***********************************
	// Delete the column and check status.
	// ***********************************
	const unsigned int	columnNumber = column.columnNumber();

	fits_delete_col(
			_fptr,
			columnNumber + 1,
			&status );

	fits_check(_fptr);

	ENDPROC
}


bool
HiLowFitsDatabase::hasEntry( const DataSet & /* dataSet */ )
{
	PROC
	if( !locked() )
	{
		// TODO
	} // if
	RETURN true;
}


bool
HiLowFitsDatabase::hasEntry( const Block & block )
{
	PROC

	if( locked() )
	{
		RETURN true;
	} // if

	switch( block.type() )
	{
		case Block::TableT :
		{
			const Table * table = block.parent()->table( block.name() );
			RETURN hasEntry( *table );
		}
			break;
		case Block::ArrayT :
		{
			const Array * array = block.parent()->array(block.name());
			RETURN hasEntry( *array );
		}
			break;
		default :
			errstr << "the block '" << block.name() << "' has an invalid type " << int( block.type() ) << Sas::error("internalError");
			break;
	} // switch

	RETURN true;
}


bool
HiLowFitsDatabase::hasEntry( const Table & /* table */ )
{
	PROC
	if( !locked() )
	{
		// TODO
	} // if
	RETURN true;
}


bool
HiLowFitsDatabase::hasEntry( const Array & /* array */ )
{
	PROC
	if( !locked() )
	{
		// TODO
	} // if
	RETURN true;
}


bool
HiLowFitsDatabase::hasEntry( const Column & /* column */ )
{
	PROC
	if( !locked() )
	{
		// TODO
	} // if
	RETURN true;
}


void
HiLowFitsDatabase::insertRows(
		const Table &	table,
		unsigned long	pos,
		unsigned long	count )
{
	PROC

	moveToExtension( table );

	// **********************
	// Insert the table rows.
	// **********************
	fits_insert_rows(
			_fptr,
			pos,
			count,
			&status );
	fits_check(_fptr);

	ENDPROC
}


void
HiLowFitsDatabase::deleteRows(
		const Table &	table,
		unsigned long	from,
		unsigned long	count )
{
	PROC

	moveToExtension( table );

	// **********************
	// Delete the table rows.
	// **********************
	fits_delete_rows(
			_fptr,
			from + 1,
			count,
			&status );
	fits_check(_fptr);

	ENDPROC
}


template<class T>
static void
copyFixedLengthColumnRows(
	fitsfile *	_fptr,
	const Column &	column,
	unsigned long	f,
	unsigned long	t,
	unsigned long	c )
{
	PROC

	const unsigned		n = column.elements();
	T *			values = new T[ n * c ];
	const unsigned int	columnNr = column.columnNumber() + 1;
	int			type = toFITSDataType( column.dataType() );

	fits_read_col(
			_fptr,
			type,
			columnNr,
			f + 1,
			1,
			n * c,
			0,
			values,
			0,
			&status );

	fits_write_col(
			_fptr,
			type,
			columnNr,
			t + 1,
			1,
			n * c,
			values,
			&status );

	fits_check( _fptr );

	delete[] values;
	ENDPROC
}


static void
copyFixedLengthColumnStringRows(
	fitsfile *	_fptr,
	const Column &	column,
	unsigned long	f,
	unsigned long	t,
	unsigned long	c )
{
	PROC

	const unsigned		n = column.elements();
	char*			values = new char[ n + 1  ];
	const unsigned int	columnNr = column.columnNumber() + 1;
	int			type = toFITSDataType( column.dataType() );
	int			anynul;

	if( f < t )
	{
		for( unsigned long i = 0; i < c; ++i )
		{
			fits_read_col(
				_fptr,
				type,
				columnNr,
				f + c - i,
				1,
				1,
				0,
				&values,
				&anynul,
				&status );

			fits_write_col(
				_fptr,
				type,
				columnNr,
				t + c - i,
				1,
				1,
				&values,
				&status );
		} // for
	}
	else
	{
		for( unsigned long i = 0; i < c; ++i )
		{
			fits_read_col(
				_fptr,
				type,
				columnNr,
				f + 1 + i,
				1,
				1,
				0,
				&values,
				&anynul,
				&status );

			fits_write_col(
				_fptr,
				type,
				columnNr,
				t + 1 + i,
				1,
				1,
				&values,
				&status );
		} // for
	} // if

	delete[] values;
	fits_check( _fptr );

	ENDPROC
}


static void
copyFixedLengthColumnRows(
	fitsfile *	_fptr,
	const Column &	column,
	unsigned long	f,
	unsigned long	t,
	unsigned long	c )
{
	PROC

	switch( column.dataType() )
	{
		case Column::Bool :
			copyFixedLengthColumnRows<bool8>( _fptr, column, f, t, c );
			break;
		case Column::Int8 :
			copyFixedLengthColumnRows<int8>( _fptr, column, f, t, c );
			break;
		case Column::Uint16 :
			copyFixedLengthColumnRows<uint16>( _fptr, column, f, t, c );
			break;
		case Column::Int16 :
			copyFixedLengthColumnRows<int16>( _fptr, column, f, t, c );
			break;
		case Column::Uint32 :
			copyFixedLengthColumnRows<uint32>( _fptr, column, f, t, c );
			break;
		case Column::Int32 :
			copyFixedLengthColumnRows<int32>( _fptr, column, f, t, c );
			break;
		case Column::Real32 :
			copyFixedLengthColumnRows<real32>( _fptr, column, f, t, c );
			break;
		case Column::Real64 :
			copyFixedLengthColumnRows<real64>( _fptr, column, f, t, c );
			break;
		case Column::DString :
			copyFixedLengthColumnStringRows( _fptr, column, f, t, c );
			break;
		default :
			cerr << "Invalid column type" << Sas::error( "dalInternalError" );
			break;
	} // switch
	ENDPROC
}



static void
copyVariableLengthColumnRows(
	fitsfile *	_fptr,
	const Column &	column,
	unsigned long	f,
	unsigned long	t,
	unsigned long	c )
{
	PROC

		// TODO
	errstr << "Not implemented" << Sas::error( "NotImplemented" ) << _fptr << column.name() << f << t << c;
	ENDPROC
}


void
HiLowFitsDatabase::copyRows(
		const Column &	column,
		unsigned long	from,
		unsigned long	to,
		unsigned long	count )
{
	PROC

	moveToExtension( *column.parent() );

	// *********************
	// Copy the column rows.
	// *********************
	if( column.cellType() == Column::Fixed )
	{
		::copyFixedLengthColumnRows( _fptr, column, from, to, count );
	}
	else
	{
		::copyVariableLengthColumnRows( _fptr, column, from, to, count );
	} // if

	ENDPROC
}


template<class T>
static void
dumpData(
		fitsfile*	_fptr,
		const Column & 	column,
		T * 		values,
		unsigned long 	from,
		unsigned long 	count )
{
	PROC

	const unsigned int	columnNumber = column.columnNumber();

	// ***************
	// General Message
	// ***************
	if( errHandler.verbosity(Msg::App) >= Msg::Verbose )
	{
		errstr << "Writing data for column with qualified name " << column.qualifiedName() << " (column number is " << columnNumber << ") from row " << from << " to row " << from + count - 1 << Sas::message( Msg::Core, Msg::Verbose );
	} // if



	// **********************
	// Write the column data.
	// **********************
	fits_write_col(
			_fptr,
			toFITSDataType( column.dataType() ),
			columnNumber + 1,
			from + 1,
			1,
			//count * column.elements(),
			count,
			values,
			&status );
	fits_check(_fptr);

	ENDPROC
}


template<class T>
static void
dumpData(
		fitsfile*	_fptr,
		const Column & 	column,
		unsigned long	rowNumber,
		T * 		values,
		unsigned long 	from,
		unsigned long 	count )
{
	PROC

	const unsigned int	columnNumber = column.columnNumber();

	// ***************
	// General Message
	// ***************
	if( errHandler.verbosity(Msg::App) >= Msg::Verbose )
	{
		errstr << "Writing data for column with qualified name " << column.qualifiedName() << " (column number is " << columnNumber << ") from row " << from << " beginning with element " << from << " until element " << from + count - 1 << Sas::message( Msg::Core, Msg::Verbose );
	} // if


	// **********************
	// Write the column data.
	// **********************
	fits_write_col(
			_fptr,
			toFITSDataType( column.dataType() ),
			columnNumber + 1,
			rowNumber + 1,
			from + 1,
			count,
			values,
			&status );

	// FIXME: fits_set_hdustruc is deprecated. Also, it is not
	// clear why it is needed here, and it should have been
	// transparently called by fits_write_col. SOC-SPR-2390

//		fits_set_hdustruc(_fptr, &status);

	fits_check(_fptr);

	ENDPROC
}


void 
HiLowFitsDatabase::dumpData(
		const Column & 	column,
		char ** 	values,
		unsigned long 	from,
		unsigned long 	count )
{
	PROC

	moveToExtension( *column.parent() );

	const unsigned int	columnNumber = column.columnNumber();

	// ***************
	// General Message
	// ***************
	if( errHandler.verbosity(Msg::App) >= Msg::Verbose )
	{
		errstr << "Writing data for column with qualified name " << column.qualifiedName() << " (column number is " << columnNumber << ") from row " << from << " to row " << from + count - 1 << Sas::message( Msg::Core, Msg::Verbose );
	} // if


	// ******************************************
	// Write the column data.
	// Write column one string (row) at a time.
	//
	// FITS will expect a null-terminated string,
	// so it is necessary to allocate a
	// buffer which is large enough to hold
	// the string value together with the
	// null character.
	// ******************************************
	char * theseChars = new char[ column.elements() + 1 ];

	for( unsigned int i = 0; i < count / column.elements(); ++i )
	{
		// **************************************
		// Copy the the DAL memory to the buffer,
		// and then null-terminate.
		// **************************************
	 	char * p = (*values) + i * column.elements();
		strncpy( theseChars, p, column.elements() );
		theseChars[ column.elements() ] = '\0';

		fits_write_col(
			_fptr,
			toFITSDataType( column.dataType() ),
			columnNumber + 1,
			//from / column.elements() + i + 1,
			from + i + 1,
			1,
			1,
			&theseChars,
			&status );

		fits_check(_fptr);

	} // for

	delete[] theseChars;

	ENDPROC
}


void 
HiLowFitsDatabase::dumpData(
		const Column & 	column,
		unsigned long	rowNumber,
		char ** 	values,
		unsigned long 	from,
		unsigned long 	count )
{
	PROC

	moveToExtension( *column.parent() );

	const unsigned int	columnNumber = column.columnNumber();

	// ***************
	// General Message
	// ***************
	if( errHandler.verbosity(Msg::App) >= Msg::Verbose )
	{
		errstr << "Writing data for column with qualified name " << column.qualifiedName() << " (column number is " << columnNumber << ") from row " << from << " beginning with element " << from << " until element " << from + count - 1 << Sas::message( Msg::Core, Msg::Verbose );
	} // if


	// ******************************************
	// Write the string to the fits file.
	//
	// FITS will expect a null-terminated string,
	// so it is necessary to allocate a
	// buffer which is large enough to hold
	// the string value together with the
	// null character.
	// ******************************************
	char * theseValues = new char[ count + 1 ];

	// **************************************
	// Copy the the DAL memory to the buffer,
	// and then null-terminate.
	// **************************************
	strncpy( theseValues, *values, count );
	theseValues[ count ] = '\0';

	fits_write_col(
		_fptr,
		toFITSDataType( column.dataType() ),
		columnNumber + 1,
		rowNumber + 1, 
		from + 1,
		count,
		&theseValues,
		&status );

	fits_check(_fptr);

	delete[] theseValues;

	ENDPROC
}


void 
HiLowFitsDatabase::dumpData(
		const Column & 	column,
		bool8 * 	values,
		unsigned long 	from,
		unsigned long 	count )
{
	PROC
	moveToExtension( *column.parent() );
	::dumpData( _fptr, column, values, from, count );
	ENDPROC
}


void 
HiLowFitsDatabase::dumpData(
		const Column & 	column,
		int8 * 		values,
		unsigned long 	from,
		unsigned long 	count )
{
	PROC
	moveToExtension( *column.parent() );
	::dumpData( _fptr, column, values, from, count );
	ENDPROC
}


void
HiLowFitsDatabase::dumpData(
		const Column & 	column,
		uint16 * 	values,
		unsigned long 	from,
		unsigned long 	count )
{
	PROC
	moveToExtension( *column.parent() );
	::dumpData( _fptr, column, values, from, count );
	ENDPROC
}


void
HiLowFitsDatabase::dumpData(
		const Column & 	column,
		int16 * 	values,
		unsigned long 	from,
		unsigned long 	count )
{
	PROC
	moveToExtension( *column.parent() );
	::dumpData( _fptr, column, values, from, count );
	ENDPROC
}


void
HiLowFitsDatabase::dumpData(
		const Column & 	column,
		uint32 * 	values,
		unsigned long 	from,
		unsigned long 	count )
{
	PROC
	moveToExtension( *column.parent() );
	::dumpData( _fptr, column, values, from, count );
	ENDPROC
}


void
HiLowFitsDatabase::dumpData(
		const Column & 	column,
		int32 * 	values,
		unsigned long 	from,
		unsigned long 	count )
{
	PROC
	moveToExtension( *column.parent() );
	::dumpData( _fptr, column, values, from, count );
	ENDPROC
}


void
HiLowFitsDatabase::dumpData(
		const Column & 	column,
		real32 * 	values,
		unsigned long 	from,
		unsigned long 	count )
{
	PROC
	moveToExtension( *column.parent() );
	::dumpData( _fptr, column, values, from, count );
	ENDPROC
}


void
HiLowFitsDatabase::dumpData(
		const Column & 	column,
		real64 * 	values,
		unsigned long 	from,
		unsigned long 	count )
{
	PROC
	moveToExtension( *column.parent() );
	::dumpData( _fptr, column, values, from, count );
	ENDPROC
}


void 
HiLowFitsDatabase::dumpData(
		const Column & 	column,
		unsigned long	rowNumber,
		bool8 * 	values,
		unsigned long 	from,
		unsigned long 	count )
{
	PROC
	moveToExtension( *column.parent() );
	::dumpData( _fptr, column, rowNumber, values, from, count );
	ENDPROC
}


void 
HiLowFitsDatabase::dumpData(
		const Column & 	column,
		unsigned long	rowNumber,
		int8 * 		values,
		unsigned long 	from,
		unsigned long 	count )
{
	PROC
	moveToExtension( *column.parent() );
	::dumpData( _fptr, column, rowNumber, values, from, count );
	ENDPROC
}


void
HiLowFitsDatabase::dumpData(
		const Column & 	column,
		unsigned long	rowNumber,
		uint16 * 	values,
		unsigned long 	from,
		unsigned long 	count )
{
	PROC
	moveToExtension( *column.parent() );
	::dumpData( _fptr, column, rowNumber, values, from, count );
	ENDPROC
}


void
HiLowFitsDatabase::dumpData(
		const Column & 	column,
		unsigned long	rowNumber,
		int16 * 	values,
		unsigned long 	from,
		unsigned long 	count )
{
	PROC
	moveToExtension( *column.parent() );
	::dumpData( _fptr, column, rowNumber, values, from, count );
	ENDPROC
}


void
HiLowFitsDatabase::dumpData(
		const Column & 	column,
		unsigned long	rowNumber,
		uint32 * 	values,
		unsigned long 	from,
		unsigned long 	count )
{
	PROC
	moveToExtension( *column.parent() );
	::dumpData( _fptr, column, rowNumber, values, from, count );
	ENDPROC
}


void
HiLowFitsDatabase::dumpData(
		const Column & 	column,
		unsigned long	rowNumber,
		int32 * 	values,
		unsigned long 	from,
		unsigned long 	count )
{
	PROC
	moveToExtension( *column.parent() );
	::dumpData( _fptr, column, rowNumber, values, from, count );
	ENDPROC
}


void
HiLowFitsDatabase::dumpData(
		const Column & 	column,
		unsigned long	rowNumber,
		real32 * 	values,
		unsigned long 	from,
		unsigned long 	count )
{
	PROC
	moveToExtension( *column.parent() );
	::dumpData( _fptr, column, rowNumber, values, from, count );
	ENDPROC
}


void
HiLowFitsDatabase::dumpData(
		const Column & 	column,
		unsigned long	rowNumber,
		real64 * 	values,
		unsigned long 	from,
		unsigned long 	count )
{
	PROC
	moveToExtension( *column.parent() );
	::dumpData( _fptr, column, rowNumber, values, from, count );
	ENDPROC
}


template<class T>
static void
loadData(
		fitsfile*	_fptr,
		const Column & 	column,
		T * 		values,
		unsigned long 	from,
		unsigned long 	count )
{
	PROC

	const unsigned int	columnNumber = column.columnNumber();
	int			anynul;

	// ***************
	// General Message
	// ***************
	if( errHandler.verbosity(Msg::App) >= Msg::Verbose )
	{
		errstr << "Reading data for column with qualified name " << column.qualifiedName() << " (column number is " << columnNumber << ") from row " << from << " to row " << from + count - 1 << Sas::message( Msg::Core, Msg::Verbose );
	} // if


	// *********************
	// Read the column data.
	// *********************
	fits_read_col(
			_fptr,
			toFITSDataType( column.dataType() ),
			columnNumber + 1,
			from + 1,
			1,
			count,
			0,
			values,
			&anynul,
			&status );
	fits_check(_fptr);

	ENDPROC
}


template<class T>
static void
loadData(
		fitsfile*	_fptr,
		const Column & 	column,
		unsigned long	rowNumber,
		T * 		values,
		unsigned long 	from,
		unsigned long 	count )
{
	PROC

	const unsigned int	columnNumber = column.columnNumber();
	int			anynul;

	// ***************
	// General Message
	// ***************
	if( errHandler.verbosity(Msg::App) >= Msg::Verbose )
	{
		errstr << "Reading data for column with qualified name " << column.qualifiedName() << " (column number is " << columnNumber << ") from row " << from << " beginning with element " << from << " until element " << from + count - 1 << Sas::message( Msg::Core, Msg::Verbose );
	} // if


	// *********************
	// Read the column data.
	// *********************
	fits_read_col(
			_fptr,
			toFITSDataType( column.dataType() ),
			columnNumber + 1,
			rowNumber + 1,
			from + 1,
			count,
			0,
			values,
			&anynul,
			&status );
	fits_check(_fptr);

	ENDPROC
}


void
HiLowFitsDatabase::loadData(
		const Column & 	column,
		char ** 	values,
		unsigned long 	from,
		unsigned long 	count )
{
	PROC

	moveToExtension( *column.parent() );

	const unsigned int	columnNumber = column.columnNumber();
	int			anynul;

	// ***************
	// General Message
	// ***************
	if( errHandler.verbosity(Msg::App) >= Msg::Verbose )
	{
		errstr << "Reading data for column with qualified name " << column.qualifiedName() << " (column number is " << columnNumber << ") from row " << from << " to row " << from + count - 1 << Sas::message( Msg::Core, Msg::Verbose );
	} // if


	// ******************************************
	// Read the column data.
	// Read column one string (row) at a time.
	//
	// FITS will null-terminate strings, so it is
	// necessary to allocate a buffer which is
	// large enough to hold the string value
	// together with the null character.
	// ******************************************
	char * theseChars = new char[ column.elements() + 1 ];

	for( unsigned int i = 0; i < count / column.elements(); ++i )
	{
		fits_read_col(
			_fptr,
			toFITSDataType( column.dataType() ),
			columnNumber + 1,
			//from / column.elements() + i + 1,
			from + i + 1,
			1,
			1,
			0,
			&theseChars,
			&anynul,
			&status );

		fits_check( _fptr );

		// **********************************
		// Copy the buffer to the DAL memory.
		// **********************************
	 	char * p = (*values) + i * column.elements();
		strncpy( p, theseChars, column.elements() );

	} // for

	delete[] theseChars;

	ENDPROC
}


void
HiLowFitsDatabase::loadData(
		const Column & 	column,
		unsigned long	rowNumber,
		char ** 	values,
		unsigned long 	from,
		unsigned long 	count )
{
	PROC

	moveToExtension( *column.parent() );

	const unsigned int	columnNumber = column.columnNumber();
	int			anynul;

	// ***************
	// General Message
	// ***************
	if( errHandler.verbosity(Msg::App) >= Msg::Verbose )
	{
		errstr << "Reading data for column with qualified name " << column.qualifiedName() << " (column number is " << columnNumber << ") from row " << from << " beginning with element " << from << " until element " << from + count - 1 << Sas::message( Msg::Core, Msg::Verbose );
	} // if


	// ***********************************************
	// Get the fits variable length column descriptor,
	// in order to determine how many elements are
	// contained in the current row.
	// Then check status.
	// ***********************************************
	long 	repeat;
	long 	offset;

	fits_read_descript(
			_fptr,
			columnNumber + 1,
			rowNumber + 1,
			&repeat,
			&offset,
			&status );

	fits_check(_fptr);

	// ******************************************
	// Read the column data.
	// Read column string from the row.
	//
	// FITS will null-terminate strings, so it is
	// necessary to allocate a buffer which is
	// large enough to hold the string value
	// together with the null character.
	// ******************************************
	char *	theseValues = new char[ repeat + 1 ];

	fits_read_col(
		_fptr,
		toFITSDataType( column.dataType() ),
		columnNumber + 1,
		rowNumber + 1,
		from + 1,	// ignored
		count,		// ignored
		0,
		&theseValues,
		&anynul,
		&status );

	fits_check( _fptr );

	// **********************************
	// Copy the buffer to the DAL memory.
	// **********************************
	strncpy( *values, theseValues, repeat );

	delete[] theseValues;

	ENDPROC
}


void
HiLowFitsDatabase::loadData(
		const Column & 	column,
		bool8 * 	values,
		unsigned long 	from,
		unsigned long 	count )
{
	PROC
	moveToExtension( *column.parent() );
	::loadData( _fptr, column, values, from, count );
	ENDPROC
}


void
HiLowFitsDatabase::loadData(
		const Column & 	column,
		int8 * 		values,
		unsigned long 	from,
		unsigned long 	count )
{
	PROC
	moveToExtension( *column.parent() );
	::loadData( _fptr, column, values, from, count );
	ENDPROC
}


void
HiLowFitsDatabase::loadData(
		const Column & 	column,
		uint16 * 	values,
		unsigned long 	from,
		unsigned long 	count )
{
	PROC
	moveToExtension( *column.parent() );
	::loadData( _fptr, column, values, from, count );
	ENDPROC
}


void
HiLowFitsDatabase::loadData(
		const Column & 	column,
		int16 * 	values,
		unsigned long 	from,
		unsigned long 	count )
{
	PROC
	moveToExtension( *column.parent() );
	::loadData( _fptr, column, values, from, count );
	ENDPROC
}


void
HiLowFitsDatabase::loadData(
		const Column & 	column,
		uint32 * 	values,
		unsigned long 	from,
		unsigned long 	count )
{
	PROC
	moveToExtension( *column.parent() );
	::loadData( _fptr, column, values, from, count );
	ENDPROC
}


void
HiLowFitsDatabase::loadData(
		const Column & 	column,
		int32 * 	values,
		unsigned long 	from,
		unsigned long 	count )
{
	PROC
	moveToExtension( *column.parent() );
	::loadData( _fptr, column, values, from, count );
	ENDPROC
}


void
HiLowFitsDatabase::loadData(
		const Column & 	column,
		real32 * 	values,
		unsigned long 	from,
		unsigned long 	count )
{
	PROC
	moveToExtension( *column.parent() );
	::loadData( _fptr, column, values, from, count );
	ENDPROC
}


void
HiLowFitsDatabase::loadData(
		const Column & 	column,
		real64 * 	values,
		unsigned long 	from,
		unsigned long 	count )
{
	PROC
	moveToExtension( *column.parent() );
	::loadData( _fptr, column, values, from, count );
	ENDPROC
}



void
HiLowFitsDatabase::loadData(
		const Column & 	column,
		unsigned long	rowNumber,
		bool8 *		values,
		unsigned long 	from,
		unsigned long 	count )
{
	PROC
	moveToExtension( *column.parent() );
	::loadData( _fptr, column, rowNumber, values, from, count );
	ENDPROC
}


void
HiLowFitsDatabase::loadData(
		const Column & 	column,
		unsigned long	rowNumber,
		int8 * 		values,
		unsigned long 	from,
		unsigned long 	count )
{
	PROC
	moveToExtension( *column.parent() );
	::loadData( _fptr, column, rowNumber, values, from, count );
	ENDPROC
}


void
HiLowFitsDatabase::loadData(
		const Column & 	column,
		unsigned long	rowNumber,
		uint16 * 	values,
		unsigned long 	from,
		unsigned long 	count )
{
	PROC
	moveToExtension( *column.parent() );
	::loadData( _fptr, column, rowNumber, values, from, count );
	ENDPROC
}


void
HiLowFitsDatabase::loadData(
		const Column & 	column,
		unsigned long	rowNumber,
		int16 * 	values,
		unsigned long 	from,
		unsigned long 	count )
{
	PROC
	moveToExtension( *column.parent() );
	::loadData( _fptr, column, rowNumber, values, from, count );
	ENDPROC
}


void
HiLowFitsDatabase::loadData(
		const Column & 	column,
		unsigned long	rowNumber,
		uint32 * 	values,
		unsigned long 	from,
		unsigned long 	count )
{
	PROC
	moveToExtension( *column.parent() );
	::loadData( _fptr, column, rowNumber, values, from, count );
	ENDPROC
}


void
HiLowFitsDatabase::loadData(
		const Column & 	column,
		unsigned long	rowNumber,
		int32 * 	values,
		unsigned long 	from,
		unsigned long 	count )
{
	PROC
	moveToExtension( *column.parent() );
	::loadData( _fptr, column, rowNumber, values, from, count );
	ENDPROC
}


void
HiLowFitsDatabase::loadData(
		const Column & 	column,
		unsigned long	rowNumber,
		real32 * 	values,
		unsigned long 	from,
		unsigned long 	count )
{
	PROC
	moveToExtension( *column.parent() );
	::loadData( _fptr, column, rowNumber, values, from, count );
	ENDPROC
}


void
HiLowFitsDatabase::loadData(
		const Column & 	column,
		unsigned long	rowNumber,
		real64 * 	values,
		unsigned long 	from,
		unsigned long 	count )
{
	PROC
	moveToExtension( *column.parent() );
	::loadData( _fptr, column, rowNumber, values, from, count );
	ENDPROC
}


template<class T>
static void
dumpData(
		fitsfile*	_fptr,
		const Array &	array,
		T * 		values,
		unsigned long 	from,
		unsigned long 	count )
{
	PROC

	// ***************
	// General Message
	// ***************
	if( errHandler.verbosity(Msg::App) >= Msg::Verbose )
	{
		errstr << "Writing data for array with qualified name " << array.qualifiedName() << " from element " << from << " to element " << from + count - 1 << Sas::message( Msg::Core, Msg::Verbose );
	} // if


	// *********************
	// Write the array data.
	// *********************
	fits_write_img(
			_fptr,
			toFITSDataType( array.dataType() ),
			from + 1,
			//count * array.elements(),
			count,
			values,
			&status );
	fits_check(_fptr);

	ENDPROC
}


void
HiLowFitsDatabase::dumpData(
		const Array & 	array, 
		char ** 	values,
		unsigned long 	from,
		unsigned long 	count )
{
	PROC
	moveToExtension( array );
	::dumpData( _fptr, array, values, from, count );
	ENDPROC
}


void
HiLowFitsDatabase::dumpData(
		const Array & 	array, 
		bool8 * 	values,
		unsigned long 	from,
		unsigned long 	count )
{
	PROC
	moveToExtension( array );
	::dumpData( _fptr, array, values, from, count );
	ENDPROC
}


void
HiLowFitsDatabase::dumpData(
		const Array & 	array, 
		int8 * 		values,
		unsigned long 	from,
		unsigned long 	count )
{
	PROC
	moveToExtension( array );
	::dumpData( _fptr, array, values, from, count );
	ENDPROC
}


void
HiLowFitsDatabase::dumpData(
		const Array & 	array,
		uint16 * 	values,
		unsigned long 	from,
		unsigned long 	count )
{
	PROC
	moveToExtension( array );
	::dumpData( _fptr, array, values, from, count );
	ENDPROC
}


void
HiLowFitsDatabase::dumpData(
		const Array & 	array,
		int16 * 	values,
		unsigned long 	from,
		unsigned long 	count )
{
	PROC
	moveToExtension( array );
	::dumpData( _fptr, array, values, from, count );
	ENDPROC
}


void
HiLowFitsDatabase::dumpData(
		const Array & 	array,
		uint32 * 	values,
		unsigned long 	from,
		unsigned long 	count )
{
	PROC
	moveToExtension( array );
	::dumpData( _fptr, array, values, from, count );
	ENDPROC
}


void
HiLowFitsDatabase::dumpData(
		const Array & 	array,
		int32 * 	values,
		unsigned long 	from,
		unsigned long 	count )
{
	PROC
	moveToExtension( array );
	::dumpData( _fptr, array, values, from, count );
	ENDPROC
}


void
HiLowFitsDatabase::dumpData(
		const Array & 	array,
		real32 * 	values,
		unsigned long 	from,
		unsigned long 	count )
{
	PROC
	moveToExtension( array );
	::dumpData( _fptr, array, values, from, count );
	ENDPROC
}


void
HiLowFitsDatabase::dumpData(
		const Array & 	array,
		real64 * 	values,
		unsigned long 	from,
		unsigned long 	count )
{
	PROC
	moveToExtension( array );
	::dumpData( _fptr, array, values, from, count );
	ENDPROC
}


template<class T>
static void
loadData(
		fitsfile*	_fptr,
		const Array & 	array,
		T * 		values,
		unsigned long 	from,
		unsigned long 	count )
{
	PROC

	int			anynul;

	// ***************
	// General Message
	// ***************
	if( errHandler.verbosity(Msg::App) >= Msg::Verbose )
	{
		errstr << "Reading data for array with qualified name " << array.qualifiedName() << " from element " << from << " to element " << from + count - 1 << Sas::message( Msg::Core, Msg::Verbose );
	} // if


	// ********************
	// Read the image data.
	// ********************
	fits_read_img(
			_fptr,
			toFITSDataType( array.dataType() ),
			from + 1,
			count,
			0,
			values,
			&anynul,
			&status );
	fits_check(_fptr);

	ENDPROC
}


void
HiLowFitsDatabase::loadData(
		const Array & 	array,
		char ** 	values,
		unsigned long 	from,
		unsigned long 	count )
{
	PROC
	moveToExtension( array );
	::loadData( _fptr, array, values, from, count );
	ENDPROC
}


void
HiLowFitsDatabase::loadData(
		const Array & 	array,
		bool8 *		values,
		unsigned long 	from,
		unsigned long 	count )
{
	PROC
	moveToExtension( array );
	::loadData( _fptr, array, values, from, count );
	ENDPROC
}


void
HiLowFitsDatabase::loadData(
		const Array & 	array,
		int8 * 		values,
		unsigned long 	from,
		unsigned long 	count )
{
	PROC
	moveToExtension( array );
	::loadData( _fptr, array, values, from, count );
	ENDPROC
}


void
HiLowFitsDatabase::loadData(
		const Array & 	array,
		uint16 * 	values,
		unsigned long 	from,
		unsigned long 	count )
{
	PROC
	moveToExtension( array );
	::loadData( _fptr, array, values, from, count );
	ENDPROC
}


void
HiLowFitsDatabase::loadData(
		const Array & 	array,
		int16 * 	values,
		unsigned long 	from,
		unsigned long 	count )
{
	PROC
	moveToExtension( array );
	::loadData( _fptr, array, values, from, count );
	ENDPROC
}


void
HiLowFitsDatabase::loadData(
		const Array & 	array,
		uint32 * 	values,
		unsigned long 	from,
		unsigned long 	count )
{
	PROC
	moveToExtension( array );
	::loadData( _fptr, array, values, from, count );
	ENDPROC
}


void
HiLowFitsDatabase::loadData(
		const Array & 	array,
		int32 * 	values,
		unsigned long 	from,
		unsigned long 	count )
{
	PROC
	moveToExtension( array );
	::loadData( _fptr, array, values, from, count );
	ENDPROC
}


void
HiLowFitsDatabase::loadData(
		const Array & 	array,
		real32 * 	values,
		unsigned long 	from,
		unsigned long 	count )
{
	PROC
	moveToExtension( array );
	::loadData( _fptr, array, values, from, count );
	ENDPROC
}


void
HiLowFitsDatabase::loadData(
		const Array &	array,
		real64 * 	values,
		unsigned long 	from,
		unsigned long 	count )
{
	PROC
	moveToExtension( array );
	::loadData( _fptr, array, values, from, count );
	ENDPROC
}


void
HiLowFitsDatabase::openFitsFile( const Block& block )
{
	PROC
	// ***********************************************
	// If the fits file is not open, then open it.
	// The open mode is dependant on the block mode().
	// ***********************************************
	if( !_fptr )
	{
		const string		fileName = block.parent() -> name();

		if( block.mode() == Dal::Read )
		{
			fitsOpenFile(&_fptr, fileName.c_str(), READONLY, &status );
		}
		else
		{
			fitsOpenFile(&_fptr, fileName.c_str(), READWRITE, &status );
		} // if
		fits_check(_fptr);
	} // if

	ENDPROC
}


void
HiLowFitsDatabase::moveToExtension( const Table& table, MoveToMode moveToMode)
{
	PROC

	openFitsFile( table );

	// **********************************
	// Move to the appropriate extension.
	// **********************************
	int			hdutype;
	const int		hdunum = blockNumberToHdunum( table );

	fits_movabs_hdu(
			_fptr,
			(moveToMode == To) ? hdunum : hdunum - 1,
			&hdutype,
			&status );
	fits_check(_fptr);

	ENDPROC
}


void
HiLowFitsDatabase::moveToExtension( const Array& array, MoveToMode moveToMode)
{
	PROC

	openFitsFile( array );

	int hn;
	fits_get_hdu_num( _fptr, &hn );

	// **********************************
	// Move to the appropriate extension.
	// **********************************
	int			hdutype;
	const int		hdunum = blockNumberToHdunum( array );

	fits_flush_file( _fptr, &status );
	fits_check(_fptr);
	if( hn == hdunum ) ALTRETURN;

	// *********************************************
	// Check to see if this is the "PRIMARY" image.
	// This has essentially already been done in the
	// call to blockNumberToHdunum().
	// *********************************************
	if( hdunum == 1 )
	{
		// **********************************
		// Move to the appropriate extension.
		// **********************************
		fits_movabs_hdu(
			_fptr,
			1,
			&hdutype,
			&status );
		fits_check(_fptr);

	}
	else
	{
		// **********************************
		// Move to the appropriate extension.
		// **********************************
		fits_movabs_hdu(
			_fptr,
			(moveToMode == To) ? hdunum : hdunum - 1,
			&hdutype,
			&status );
		fits_check(_fptr);

	}// if

	ENDPROC
}

