using namespace std;
/*************************************************************** 
* 
* MemFitsReaderwriter.h     & XMM Science Analysis System (SAS) 
* 
* Class representing Fits File reader/writer.
* 
* (c) 1997 ESA - Written by Mark Thomas, CARA
*                and Marco Beijersbergen
* 
* $Id: MemFitsReaderWriter.cc,v 1.3 2005/06/06 09:00:10 jhoar Exp $
* 
*****************************************************************/ 


#include <string>
#include <iostream>
#include <fitsio.h>
#include <unistd.h>
#include <sstream>

#include <stringutil.h>
#include <Tracer.h>
#include <errstr.h>
#include <Dal.h>

#include "utilities.h"
#include "MemFitsReaderWriter.h"
#include "dal_utilities.h"


// ********************
// Global defined here.
// ********************

namespace {

 const string nameOfPrimaryImage = "PRIMARY";
 const string defaultArrayComment = "The name of this array";
 const string defaultTableComment = "The name of this table";


 int
writeVariableLengthColumnData(
			      fitsfile * fptr,
			      const Column * column,
			      int &  status )
{

  PROC

    if( status != 0 )
      {
	ALTRETURNV( 1 );
      } // if

  int  columnNumber = column->columnNumber();
  for( unsigned int j = 0; j < column -> rows(); ++j )
    {

      // *******************************************************
      // Get access to the cell data of the column. This will be
      // deleted after the switch.
      // *******************************************************
      CellData *data = column->cellData( j );

      switch( column -> dataType() )
	{
	case Column::Int8 :

	  fits_write_col_byt(
			     fptr,
			     columnNumber + 1,
			     j + 1,
			     1, // Needed for Binary Tables
			     data -> size(),
			     (unsigned char*)data -> int8Data(),
			     &status );
       
	  break;

	case Column::Uint16 :

	  fits_write_col_usht(
			      fptr,
			      columnNumber + 1,
			      j + 1,
			      1, // Needed for Binary Tables
			      data -> size(),
			      data -> uint16Data(),
			      &status );
      
	  break;

	case Column::Int16 :

	  fits_write_col_sht(
			     fptr,
			     columnNumber + 1,
			     j + 1,
			     1, // Needed for Binary Tables
			     data -> size(),
			     data -> int16Data(),
			     &status );
      
	  break;

	case Column::Uint32 :

	  // ==========================
	  // Use TLONG to be consistant
	  // with reading TLONGs.
	  // ==========================
	  fits_write_col_uint(
			      fptr,
			      columnNumber + 1,
			      j + 1,
			      1, // Needed for Binary Tables
			      data -> size(),
			      data -> uint32Data(),
			      &status );
      
	  break;

	case Column::Int32 :

	  // ==========================
	  // Use TLONG to be consistant
	  // with reading TLONGs.
	  // ==========================
	  fits_write_col_int(
			     fptr,
			     columnNumber + 1,
			     j + 1,
			     1, // Needed for Binary Tables
			     data -> size(),
			     data -> int32Data(),
			     &status );
      
	  break;

	case Column::Real32 :

	  fits_write_col_flt(
			     fptr,
			     columnNumber + 1,
			     j + 1,
			     1, // Needed for Binary Tables
			     data -> size(),
			     data -> real32Data(),
			     &status );
      
	  break;

	case Column::Real64 :

	  fits_write_col_dbl(
			     fptr,
			     columnNumber + 1,
			     j + 1,
			     1, // Needed for Binary Tables
			     data -> size(),
			     data -> real64Data(),
			     &status );
      
	  break;

	case Column::Bool :

	  fits_write_col_log(
			     fptr,
			     columnNumber + 1,
			     j + 1,
			     1, // Needed for Binary Tables
			     data -> size(),
			     (char *)data -> boolData(),
			     &status );

	  break;

	case Column::DString :
	  {
	    // ******************************************
	    // Ensure that the string written to the
	    // fits file is null terminated. This
	    // is necessary in the variable column length
	    // case because cfitsio ignores the nelements
	    // parameter.
	    // ******************************************
	    char * theseChars = new char[ data -> size() + 1 ];

	    // **************************************
	    // Copy the DAL memory to the buffer, and
	    // then null-terminate.
	    // **************************************
	    strncpy( theseChars, data -> stringData(), data -> size() );
	    theseChars[ data -> size() ] = '\0';

	    fits_write_col_str(
			       fptr,
			       columnNumber + 1,
			       j + 1,
			       1, // Needed for Binary Tables
			       1,
			       &theseChars,
			       &status );

	    delete[] theseChars;


	  }
	break;

	default :

	  cerr << "could not write column '" << column->name() << "' of table " << column->parent()->name() << " to file '" << fits_file_name( fptr ) << "' since the column type " << int( column->dataType() ) << " is invalid" << Sas::error("internalError");

	  break;

	} // switch

      // Delete the data object again.
      delete data;

      // Check whether the write succeeded.
      fits_check(fptr);

    } // for

  // ******************************
  // Everything is Ok, so return 0.
  // ******************************

  RETURN 0;
}


int writeFixedLengthColumnData(fitsfile * fptr,
			       const Column * column,
			       long offset,
			       long nwrite,
			       int &  status)
{
  PROC

      int  columnNumber = column->columnNumber();

    if( status != 0 )
      {
	ALTRETURNV( 1 );
      } // if

  if(nwrite == 0)
      ALTRETURNV(0);

  unsigned rowoffset = offset;
  unsigned elementstowrite = nwrite * column->elements();
  unsigned dataoffset = rowoffset * column->elements();

  errstr << "Will write " << nwrite << " rows starting from " << rowoffset
	 << " that is to say " << elementstowrite
	 << " elements will be written starting from element " << dataoffset
	 << Sas::message( Msg::Core, Msg::Verbose );


  // ***********************************************************
  // Get access to the data of the column. This is to be deleted
  // after the switch again.
  // ***********************************************************
  ColumnData *data = column->data();

  switch( column -> dataType() )
  {
      case Column::Bool :

	  fits_write_col_log(
	      fptr,
	      columnNumber + 1,
	      1 + rowoffset,
	      1, // Needed for Binary Tables
	      elementstowrite,
	      (char *)data -> boolData() + dataoffset,
	      &status );
	  
      break;

    case Column::Int8 :

      fits_write_col_byt(
			 fptr,
			 columnNumber + 1,
			 1 + rowoffset,
			 1, // Needed for Binary Tables
			 elementstowrite,
			 (unsigned char*)data -> int8Data() + dataoffset,
			 &status );
      
      break;

    case Column::Uint16 :

      fits_write_col_usht(
			  fptr,
			  columnNumber + 1,
			  1 + rowoffset,
			  1, // Needed for Binary Tables
			  elementstowrite,
			  data -> uint16Data() + dataoffset,
			  &status );
      
      break;

    case Column::Int16 :

      fits_write_col_sht(
			 fptr,
			 columnNumber + 1,
			 1 + rowoffset,
			 1, // Needed for Binary Tables
			 elementstowrite,
			 data -> int16Data() + dataoffset,
			 &status );
      
      break;

    case Column::Uint32 :

      // ==========================
      // Use TLONG to be consistant
      // with reading TLONGs.
      // ==========================
      fits_write_col_uint(
			  fptr,
			  columnNumber + 1,
			  1 + rowoffset,
			  1, // Needed for Binary Tables
			  elementstowrite,
			  data -> uint32Data() + dataoffset,
			  &status );

      break;

    case Column::Int32 :

      // ==========================
      // Use TLONG to be consistant
      // with reading TLONGs.
      // ==========================
      fits_write_col_int(
			 fptr,
			 columnNumber + 1,
			 1 + rowoffset,
			 1, // Needed for Binary Tables
			 elementstowrite,
			 data -> int32Data() + dataoffset,
			 &status );
      
      break;

    case Column::Real32 :

      fits_write_col_flt(
			 fptr,
			 columnNumber + 1,
			 1 + rowoffset,
			 1, // Needed for Binary Tables
			 elementstowrite,
			 data -> real32Data() + dataoffset,
			 &status );

      break;

    case Column::Real64 :

      fits_write_col_dbl(
			 fptr,
			 columnNumber + 1,
			 1 + rowoffset,
			 1, // Needed for Binary Tables
			 elementstowrite,
			 data -> real64Data() + dataoffset,
			 &status );

      break;

    case Column::DString :
      {
	// ******************************************
	// Write the column data.
	// Write column one string (row) at a time.
	//
	// FITS will expect a null-terminated string,
	// so it is necessary to allocate a
	// buffer which is large enough to hold
	// the string value together with the
	// null character.
	// ******************************************
	const unsigned long n = nwrite;
	const unsigned long  len = column -> elements();
	char *    theseChars = new char[ len + 1 ];
	char*    chars = data -> stringData() + dataoffset;

	for( unsigned int i = 0; i < n; ++i )
	  {
	    // **********************************
	    // Copy the DAL memory to the buffer,
	    // and then null-terminate.
	    // **********************************
	    char * p = chars + i * len;
	    strncpy( theseChars, p, len );
	    theseChars[ len ] = '\0';

	    fits_write_col_str(
			       fptr,
			       columnNumber + 1,
			       i + 1 + rowoffset,
			       1,
			       1,
			       &theseChars,
			       &status );

	    fits_check(fptr);

	  } // for

	delete[] theseChars;
      }
    break;

    default :

      cerr << "could not write column '" << column->name() << "' of table " << column->parent()->name() << " to file '" << fits_file_name( fptr ) << " since the column type " << int( column->dataType() ) << "' is invalid" << Sas::error("internalError");

      break;

    } // switch

  // Delete the data object again.
  delete data;

  // Check whether the write succeeded.
  fits_check(fptr);

  // ******************************
  // Everything is Ok, so return 0.
  // ******************************

  RETURN 0;

}


 unsigned long _row;
 fitsfile* _fptr;



class BaseWriteRowWiser
{
public :
  BaseWriteRowWiser() {}
  virtual ~BaseWriteRowWiser() {}
  virtual void write() {};
};


template<class T>
class WriteRowWiser : public virtual BaseWriteRowWiser
{
public :
  WriteRowWiser(
		unsigned int columnNumber,
		unsigned long elements,
		T*  data,
		int(*fitsWrite)( fitsfile*,int,long,long,long,T*,int*) )
    :
    _columnNumber( columnNumber ),
    _elements( elements ),
    _data( data ),
    _fitsWrite( fitsWrite ) {
    _row = 1; }

protected :

  unsigned int _columnNumber;
  unsigned long _elements;
  T*  _data;
  int (*_fitsWrite)( fitsfile*,int,long,long,long,T*,int*);

};


template<class T>
class WriteRowWise : public virtual WriteRowWiser<T>
{
public :

  WriteRowWise(
	       unsigned int columnNumber,
	       unsigned long elements,
	       T*  data,
	       int(*fitsWrite)( fitsfile*,int,long,long,long,T*,int*) )
    :
    WriteRowWiser<T>( columnNumber, elements, data, fitsWrite ) {}

  void write();

};


template<class T>
inline void
WriteRowWise<T>::write()
{
  PROC

    _fitsWrite(
	       _fptr,
	       _columnNumber,
	       _row,
	       1, // Needed for Binary Tables
	       _elements,
	       _data,
	       &status );

  _data+= _elements;

  ENDPROC
    }


class StringWriteRowWise : public virtual WriteRowWiser<char>
{
public :

  StringWriteRowWise(
		     unsigned int columnNumber,
		     unsigned long elements,
		     char*  data )
    :
    WriteRowWiser<char>( columnNumber, elements, data, 0 ),
    _theseChars( new char[ _elements + 1 ] ) {}

  ~StringWriteRowWise() { delete[] _theseChars; }
  void write();

private :

  char *    _theseChars;
};


inline void
StringWriteRowWise::write()
{
  PROC

    // ******************************************
    // Write the column data.
    // Write column one string (row) at a time.
    //
    // FITS will expect a null-terminated string,
    // so it is necessary to allocate a
    // buffer which is large enough to hold
    // the string value together with the
    // null character.
    // ******************************************

    // **********************************
    // Copy the DAL memory to the buffer,
    // and then null-terminate.
    // **********************************
    strncpy( _theseChars, _data, _elements );
  _theseChars[ _elements ] = '\0';

  fits_write_col_str(
		     _fptr,
		     _columnNumber,
		     _row,
		     1,
		     1,
		     &_theseChars,
		     &status );

  //++_row;
  _data+= _elements;

  ENDPROC

    }


int writeColumnDataRowWise(
    fitsfile * fptr,
    const Table* table,
    int &  status )
{
  PROC

    if( status != 0 )
      {
	ALTRETURNV( 1 );
      } // if

  const unsigned int columns = table -> columns();
  BaseWriteRowWiser* writers[ columns ];

  for( unsigned int columnNumber = 0; columnNumber < columns; ++columnNumber )
    {
      const Column* column = table -> column( columnNumber );

      // ***********************
      // If variable length ....
      // ***********************
      if( column -> size() == Dal::zero )
	{
	  writers[ columnNumber ] = new BaseWriteRowWiser;
	  continue;
	} // if

      // ******************************
      // Else, must be fixed length ...
      // ******************************
      ColumnData* cd = column -> data();

      switch( column -> dataType() )
	{
	case Column::Int8 :
	  writers[ columnNumber ] = new WriteRowWise<unsigned char>(
								    columnNumber + 1,
								    column -> elements(),
								    cd -> int8Data(),
								    fits_write_col_byt );

	  break;

	case Column::Uint16 :
	  writers[ columnNumber ] = new WriteRowWise<uint16>(
							     columnNumber + 1,
							     column -> elements(),
							     cd -> uint16Data(),
							     fits_write_col_usht );

	  break;

	case Column::Int16 :
	  writers[ columnNumber ] = new WriteRowWise<int16>(
							    columnNumber + 1,
							    column -> elements(),
							    cd -> int16Data(),
							    fits_write_col_sht );

	  break;

	case Column::Uint32 :
	  writers[ columnNumber ] = new WriteRowWise<uint32>(
							     columnNumber + 1,
							     column -> elements(),
							     cd -> uint32Data(),
							     fits_write_col_uint );

	  break;

	case Column::Int32:
	  writers[ columnNumber ] = new WriteRowWise<int32>(
							    columnNumber + 1,
							    column -> elements(),
							    cd -> int32Data(),
							    fits_write_col_int );

	  break;

	case Column::Real32 :
	  writers[ columnNumber ] = new WriteRowWise<real32>(
							     columnNumber + 1,
							     column -> elements(),
							     cd -> real32Data(),
							     fits_write_col_flt );

	  break;

	case Column::Real64 :
	  writers[ columnNumber ] = new WriteRowWise<real64>(
							     columnNumber + 1,
							     column -> elements(),
							     cd -> real64Data(),
							     fits_write_col_dbl );

	  break;

	case Column::Bool :
	  writers[ columnNumber ] = new WriteRowWise<bool8>(
							    columnNumber + 1,
							    column -> elements(),
							    cd -> boolData(),
							    fits_write_col_log );

	  break;

	case Column::DString :
	  writers[ columnNumber ] = new StringWriteRowWise(
							   columnNumber + 1,
							   column -> elements(),
							   cd -> stringData() );

	  break;

	default :

	  cerr << "could not write column '" << column->name() << "' of table " << column->parent()->name() << " to file '" << fits_file_name( fptr ) << " since the column type " << int( column->dataType() ) << "' is invalid" << Sas::error("internalError");

	  break;
	} // switch

      delete cd;

    } // for


  const unsigned long rows = table -> rows();
  _fptr = fptr;

  for( _row = 1; _row<= rows; ++_row )
    {
      for( unsigned int columnNumber = 0; columnNumber < columns; ++columnNumber )
	{
	  writers[ columnNumber ] -> write();
	} // for
    } // for

  // Check whether the write succeeded.
  fits_check(fptr);

  for( unsigned int columnNumber = 0; columnNumber < columns; ++columnNumber )
    {
      delete writers[ columnNumber ];
    } // for

  // ******************************
  // Everything is Ok, so return 0.
  // ******************************

  RETURN 0;

}

};

int
MemFitsReaderWriter::writeColumns(
				  fitsfile * fptr,
				  const Table & table,
				  int &  status ) const
{
  PROC

    if( status != 0 )
      {
	ALTRETURNV( 1 );
      } // if

//   static bool firstTime = true;
//   static bool columnWiseWrite = true;

//   if( firstTime )
//     {
//       columnWiseWrite = getenv( "SAS_COLUMN_WISE" );
//       firstTime = false;
//     } // if

  // Sort columns by cell type: fixed length or variable length
  std::vector<const Column*> fixedlcolumns;
  std::vector<const Column*> variablelcolumns;

  // **********************
  // For each Table Column.
  // **********************
  for( unsigned int j = 0; j < table.columns(); ++j ){
      const Column *  column = table.column( j );
      
      
      if(column->cellType() == Column::Fixed)
	  fixedlcolumns.push_back(column);
      else
	  variablelcolumns.push_back(column);

      // =============================
      // write Column Scaling Factors.
      // =============================
      writeColumnScalingFactors( fptr, column, j, status );

      // =========================
      // Write the column comment.
      // =========================
      writeColumnComment( fptr, column -> label(), j, status );

      // ==========
      // write tdim
      // ==========
      if( column -> dimensions() > 1 )
	//if( column -> dimensions() > 0 && column -> elements() > 1 )
	{
	  fits_write_tdim(
			  fptr,
			  j + 1,
			  column -> dimensions(),
			  (long *)Mvec<unsigned long>( column -> size() ).data(),
			  &status );
	} // if

      // ========================
      // Write column null value.
      // ========================
      if( column -> nullDefined() )
	{
	  writeColumnNullValue( fptr, column, status );
	} // if

      // ========================
      // Write Column Attributes.
      // ========================
      writeHeader(
		  fptr,
		  *column,
		  toString( j + 1 ),
		  &FitsReaderWriter::columnAttributeOK,
		  status );

  } // for


  long nopt = 0;
  fits_get_rowsize(fptr, &nopt, &status);
  unsigned long chunk = static_cast<unsigned long>(nopt);
   if(chunk == 0 || chunk > table.rows())
	chunk = table.rows();
//     errstr << "Optimal number of rows to write for this table: "
// 	   << chunk << " (nrows = " << table.rows() << ")"
// 	   << Sas::message( Msg::Core, Msg::Verbose);
    
    unsigned nchunks = chunk == 0 ? 0 : table.rows() / chunk;
    unsigned got = 0;
    for(unsigned K = 0; K <= nchunks; ++K){
	long offset = K*chunk;
	long nwrite = (K != nchunks) ? chunk : (table.rows() - nchunks*chunk);
	got += nwrite;
	for(unsigned long k = 0; k < fixedlcolumns.size(); ++k){
	    writeFixedLengthColumnData( fptr, fixedlcolumns[k],
					offset, nwrite, status);
	}
    }

    for(unsigned long k = 0; k < variablelcolumns.size(); ++k){
	writeVariableLengthColumnData( fptr, variablelcolumns[k], status );
    }

  RETURN 0;

}


int
MemFitsReaderWriter::writeTable( 
				fitsfile * fptr,
				const Table & table,
				int &  status ) const
{
  PROC

    if( status != 0 )
      {
	ALTRETURNV( 1 );
      } // if


  // ======================================================
  // Initialise the array of column names, types and units.
  // ======================================================
  const unsigned int tfields = table.columns();
  const char *  ttype[ tfields + 1 ];
  const char *  tform[ tfields + 1 ];
  const char *  tunit[ tfields + 1 ];

  /*
	JBa: fixed danger zone!
	Use a local copy for the format, as stringstream.str().c_str() descopes (!)
   */
  string stform[ tfields ];

  for( unsigned j = 0; j < tfields; j++ )
    {
      const Column * column = table.column( j );
      stringstream format;

      const long repeat = column -> elements();

      // -------------
      // Column Names.
      // -------------
      ttype[ j ] = column -> name().c_str();

      // ------------------
      // Column Data Types.
      // ------------------

      // ****************************************
      // Check for variable length column; if not
      // then check for vector column.
      // ****************************************
      if( repeat == 0 ) format << "1P";
      else if (repeat > 1) format << repeat;

      switch( column -> dataType() )
	{
	case Column::Int8 :

	  format << "B";

	  break;

	case Column::Uint16 :

	  format << "U";

	  break;

	case Column::Int16 :

	  format << "I";

	  break;

	case Column::Uint32 :

	  format << "V";

	  break;

	case Column::Int32 :

	  format << "J";

	  break;

	case Column::Real32 :

	  format << "E";

	  break;

	case Column::Real64 :

	  format << "D";

	  break;

	case Column::Bool :

	  format << "L";

	  break;

	case Column::DString :

	  format << "A";

	  break;

	default :

	  errstr << "could not write column '" << column->name() << "' of table '" << column->parent()->name() << "' since it has an invalid data type " << int( column->dataType()) << Sas::error("internalError");

	  break;

	} // switch

      stform[ j ] = format.str();

      // -------------
      // Column units.
      // -------------
      tunit[ j ] = column -> units().c_str();

    } /* for */

  // FITS strings BEGIN
  for( unsigned j = 0; j < tfields; j++ ) tform[j]=stform[j].c_str();
  ttype[ tfields ] = (const char *)0;
  tform[ tfields ] = (const char *)0;
  tunit[ tfields ] = (const char *)0;
  // FITS strings END

  // ************************
  // Create and append table.
  // ************************
  if( getenv( "SAS_ZERO_TABLE_ROWS" ) )
    {
      fits_create_tbl(
		      fptr,
		      // ASCII_TBL,
		      BINARY_TBL,
		      0,
		      tfields,
		      (char **)ttype,
		      (char **)tform,
		      (char **)tunit,
		      "",
		      &status );
    }
  else
    {
      fits_create_tbl(
		      fptr,
		      // ASCII_TBL,
		      BINARY_TBL,
		      table.rows(),
		      tfields,
		      (char **)ttype,
		      (char **)tform,
		      (char **)tunit,
		      "",
		      &status );
    } // if

  // ***************************************************
  // Set the table name. This is achieved by setting the
  // FITS file keyword EXTNAME.
  // ***************************************************

  fits_write_key_str(
		     fptr,
		     "EXTNAME", // This is how to set the name.
		     (char *)table.name().c_str(),
		     table.label() == "" ? (char*)defaultTableComment.c_str() : (char *)table.label().c_str(),
		     &status );

  fits_check(fptr);

  // *******************
  // write table header.
  // *******************
  writeHeader(
	      fptr,
	      (const Attributable&)table,
	      "",
	      &FitsReaderWriter::tableAttributeOK,
	      status );

  // ************************
  // Write the table columns.
  // ************************
  writeColumns( fptr, table, status );

  // ******************************
  // Everything is Ok, so return 0.
  // ******************************
  RETURN 0;

}


static int
writeArrayData(
	       fitsfile * fptr,
	       const Array & array,
	       int &  status )
{
  PROC

    if( status != 0 )
      {
	ALTRETURNV( 1 );
      } // if

  // **********************************
  // Exit if there is no data to write.
  // **********************************
  if( numberOfElements( array.size() ) == 0 )
    {
      ALTRETURNV( 0 );
    } // if

  // ************************************
  // Get a data object to access the data
  // ************************************
  ArrayData* data = array.data();

  switch( array.dimensions() )
    {
    case 1 :

      switch( array.dataType() )
	{
	case Array::Int8 :

	  fits_write_img_byt(
			     fptr,
			     0,
			     1,
			     array.size()[ 0 ],
			     (unsigned char *)data -> int8Data(),
			     &status );

	  break;

	case Array::Uint16 :

	  fits_write_img_usht(
			      fptr,
			      0,
			      1,
			      array.size()[ 0 ],
			      data -> uint16Data(),
			      &status );

	  break;

	case Array::Int16 :

	  fits_write_img_sht(
			     fptr,
			     0,
			     1,
			     array.size()[ 0 ],
			     data -> int16Data(),
			     &status );

	  break;

	case Array::Uint32 :

	  fits_write_img_uint(
			      fptr,
			      0,
			      1,
			      array.size()[ 0 ],
			      data -> uint32Data(),
			      &status );

	  break;

	case Array::Int32 :

	  fits_write_img_int(
			     fptr,
			     0,
			     1,
			     array.size()[ 0 ],
			     data -> int32Data(),
			     &status );

	  break;

	case Array::Real32 :

	  fits_write_img_flt(
			     fptr,
			     0,
			     1,
			     array.size()[ 0 ],
			     data -> real32Data(),
			     &status );

	  break;

	case Array::Real64 :

	  fits_write_img_dbl(
			     fptr,
			     0,
			     1,
			     array.size()[ 0 ],
			     data -> real64Data(),
			     &status );

	  break;

	default :

	  errstr << "could not write array '" << array.name() << "' to fits file '" << fits_file_name( fptr ) << "' since it has an invalid data type " << int( array.dataType() ) << Sas::error("internalError");

	  break;

	} //switch

      break;

    case 2 :

      switch( array.dataType() )
	{
	case Array::Int8 :

	  fits_write_2d_byt(
			    fptr,
			    0,
			    array.size()[ 1 ],
			    array.size()[ 1 ],
			    array.size()[ 0 ],
			    (unsigned char *)data -> int8Data(),
			    &status );

	  break;

	case Array::Uint16 :

	  fits_write_2d_usht(
			     fptr,
			     0,
			     array.size()[ 1 ],
			     array.size()[ 1 ],
			     array.size()[ 0 ],
			     data -> uint16Data(),
			     &status );

	  break;

	case Array::Int16 :

	  fits_write_2d_sht(
			    fptr,
			    0,
			    array.size()[ 1 ],
			    array.size()[ 1 ],
			    array.size()[ 0 ],
			    data -> int16Data(),
			    &status );

	  break;

	case Array::Uint32 :

	  fits_write_2d_uint(
			     fptr,
			     0,
			     array.size()[ 1 ],
			     array.size()[ 1 ],
			     array.size()[ 0 ],
			     data -> uint32Data(),
			     &status );

	  break;

	case Array::Int32 :

	  fits_write_2d_int(
			    fptr,
			    0,
			    array.size()[ 1 ],
			    array.size()[ 1 ],
			    array.size()[ 0 ],
			    data -> int32Data(),
			    &status );

	  break;

	case Array::Real32 :

	  fits_write_2d_flt(
			    fptr,
			    0,
			    array.size()[ 1 ],
			    array.size()[ 1 ],
			    array.size()[ 0 ],
			    data -> real32Data(),
			    &status );

	  break;

	case Array::Real64 :

	  fits_write_2d_dbl(
			    fptr,
			    0,
			    array.size()[ 1 ],
			    array.size()[ 1 ],
			    array.size()[ 0 ],
			    data -> real64Data(),
			    &status );

	  break;

	default :

	  errstr << "could not write array '" << array.name() << "' to fits file '" << fits_file_name( fptr ) << "' since it has an invalid data type " << int( array.dataType() ) << Sas::error("internalError");

	  break;

	} //switch

      break;


    case 3 :

      switch( array.dataType() )
	{
	case Array::Int8 :

	  fits_write_3d_byt(
			    fptr,
			    0,
			    array.size()[ 2 ],
			    array.size()[ 1 ],
			    array.size()[ 2 ],
			    array.size()[ 1 ],
			    array.size()[ 0 ],
			    (unsigned char *)data -> int8Data(),
			    &status );

	  break;

	case Array::Uint16 :

	  fits_write_3d_usht(
			     fptr,
			     0,
			     array.size()[ 2 ],
			     array.size()[ 1 ],
			     array.size()[ 2 ],
			     array.size()[ 1 ],
			     array.size()[ 0 ],
			     data -> uint16Data(),
			     &status );

	  break;

	case Array::Int16 :

	  fits_write_3d_sht(
			    fptr,
			    0,
			    array.size()[ 2 ],
			    array.size()[ 1 ],
			    array.size()[ 2 ],
			    array.size()[ 1 ],
			    array.size()[ 0 ],
			    data -> int16Data(),
			    &status );

	  break;

	case Array::Uint32 :

	  fits_write_3d_uint(
			     fptr,
			     0,
			     array.size()[ 2 ],
			     array.size()[ 1 ],
			     array.size()[ 2 ],
			     array.size()[ 1 ],
			     array.size()[ 0 ],
			     data -> uint32Data(),
			     &status );

	  break;

	case Array::Int32 :

	  fits_write_3d_int(
			    fptr,
			    0,
			    array.size()[ 2 ],
			    array.size()[ 1 ],
			    array.size()[ 2 ],
			    array.size()[ 1 ],
			    array.size()[ 0 ],
			    data -> int32Data(),
			    &status );

	  break;

	case Array::Real32 :

	  fits_write_3d_flt(
			    fptr,
			    0,
			    array.size()[ 2 ],
			    array.size()[ 1 ],
			    array.size()[ 2 ],
			    array.size()[ 1 ],
			    array.size()[ 0 ],
			    data -> real32Data(),
			    &status );

	  break;

	case Array::Real64 :

	  fits_write_3d_dbl(
			    fptr,
			    0,
			    array.size()[ 2 ],
			    array.size()[ 1 ],
			    array.size()[ 2 ],
			    array.size()[ 1 ],
			    array.size()[ 0 ],
			    data -> real64Data(),
			    &status );

	  break;

	default :

	  errstr << "could not write array '" << array.name() << "' to fits file '" << fits_file_name( fptr ) << "' since it has an invalid data type " << int( array.dataType() ) << Sas::error("internalError");

	  break;

	} //switch

      break;

    default :

      errstr << "could not write array " << array.name() << " to fits file " << fits_file_name( fptr ) << " since its dimension " << array.dimensions() << " is not supported" << Sas::error("wrongArrayDimensions");

      break;

    } // switch

  // ************************************
  // Delete the data object again
  // ************************************
  delete data;

  fits_check(fptr);

  // ******************************
  // Everything is Ok, so return 0.
  // ******************************
  RETURN 0;

}


int
MemFitsReaderWriter::writeArray(
				fitsfile * fptr,
				const Array & array,
				int &  status ) const
{
  PROC

    if( status != 0 )
      {
	ALTRETURNV( 1 );
      } // if

  // **********************************************************
  // Revert the indices since FITS uses the Fortran convention,
  // i.e. the first index runs fastest.
  // **********************************************************
  unsigned dimensions = array.dimensions();
  long size[dimensions];
  for (unsigned i = 0; i < dimensions; ++i)
    size[i] = array.size()[dimensions-1-i];

  // *************
  // Create Image.
  // *************
  fits_create_img(
		  fptr,
		  toFITSBitpix( array.dataType() ),
		  array.dimensions(),
		  size,
		  &status );

  // ***************************************************
  // Set the array name. This is achieved by setting the
  // FITS file keyword EXTNAME.
  // ***************************************************
  //fits_write_key_str(
  //  fptr,
  //  "EXTNAME", // This is how to set the name.
  //  (char *)array.name().c_str(),
  //  (char *)array.label().c_str(),
  //  &status );

  //fits_check(fptr);

  // ********************
  // Set the array units.
  // ********************
  writeArrayUnits( fptr, array, status );


  // **********************************
  // write the array Scaling factors.
  // **********************************
  if( array.scaled() )
    {
      double  value;

      value = array.scaleZero();

      fits_write_key(
		     fptr,
		     TDOUBLE,
		     "BZERO",
		     &value,
		     "",
		     &status );

      value = array.scaleFactor();

      fits_write_key(
		     fptr,
		     TDOUBLE,
		     "BSCALE",
		     &value,
		     "",
		     &status );

      // ***********************************
      // Disable the FITS automatic scaling.
      // ***********************************
      fits_set_bscale(
		      fptr,
		      1.0,
		      0.0,
		      &status );

      fits_check(fptr);

    } // if

  // ***********************
  // Write array null value.
  // ***********************
  if( array.nullDefined() )
    {
      writeArrayNullValue( fptr, &array, status );
    } // if

  //writeHeader( fptr, (const Attributable&)array, "", 0, status );

  writeArrayData( fptr, array, status );

  // ******************************
  // Everything is Ok, so return 0.
  // ******************************
  RETURN 0;

}



int
MemFitsReaderWriter::readVariableLengthColumnData(
						  fitsfile * fptr,
						  Column * column,
						  int &  status ) const
{
  PROC

      unsigned long columnNumber = column->columnNumber();

      if( status != 0 )
      {
	  ALTRETURNV( 1 );
      } // if

  for( unsigned int i = 0; i < column -> rows(); ++i )
    {
      int   anynul;

      // ***********************************************
      // Get the fits variable length column descriptor,
      // in order to determine how many elements are
      // contained in the current row.
      // Then check status.
      // ***********************************************
      long  repeat;
      long  offset;

      fits_read_descript(
			 fptr,
			 columnNumber + 1,
			 i + 1,
			 &repeat,
			 &offset,
			 &status );

      fits_check(fptr);

      // **************************************************
      // Get a column CellData descriptor, set its size and
      // read the data from the fits file.
      // As usual the CellData descriptor object must be
      // deleted when it is no longer needed.
      // **************************************************
      CellData * c = column -> cellData( i );
      c -> size( repeat );

      switch( column -> dataType() )
	{

	case Column::Bool :

	  fits_read_col(
			fptr,
			toFITSDataType( column -> dataType() ),
			columnNumber + 1,
			i + 1,
			1,
			c -> size(),
			0,
			c -> boolData(),
			&anynul,
			&status );

	  break;

	case Column::Int8 :

	  fits_read_col(
			fptr,
			toFITSDataType( column -> dataType() ),
			columnNumber + 1,
			i + 1,
			1,
			c -> size(),
			0,
			c -> int8Data(),
			&anynul,
			&status );

	  break;

	case Column::Uint16 :

	  fits_read_col(
			fptr,
			toFITSDataType( column -> dataType() ),
			columnNumber + 1,
			i + 1,
			1,
			c -> size(),
			0,
			c -> uint16Data(),
			&anynul,
			&status );

	  break;

	case Column::Int16 :

	  fits_read_col(
			fptr,
			toFITSDataType( column -> dataType() ),
			columnNumber + 1,
			i + 1,
			1,
			c -> size(),
			0,
			c -> int16Data(),
			&anynul,
			&status );

	  break;

	case Column::Uint32 :

	  fits_read_col(
			fptr,
			toFITSDataType( column -> dataType() ),
			columnNumber + 1,
			i + 1,
			1,
			c -> size(),
			0,
			c -> uint32Data(),
			&anynul,
			&status );

	  break;

	case Column::Int32 :

	  fits_read_col(
			fptr,
			toFITSDataType( column -> dataType() ),
			columnNumber + 1,
			i + 1,
			1,
			c -> size(),
			0,
			c -> int32Data(),
			&anynul,
			&status );

	  break;

	case Column::Real32 :

	  fits_read_col(
			fptr,
			toFITSDataType( column -> dataType() ),
			columnNumber + 1,
			i + 1,
			1,
			c -> size(),
			0,
			c -> real32Data(),
			&anynul,
			&status );

	  break;

	case Column::Real64 :

	  fits_read_col(
			fptr,
			toFITSDataType( column -> dataType() ),
			columnNumber + 1,
			i + 1,
			1,
			c -> size(),
			0,
			c -> real64Data(),
			&anynul,
			&status );

	  break;

	case Column::DString :
	  {
	    // ***************************************
	    // FITS will null-terminate the string,
	    // so it is necessary to allocate
	    // a buffer large enough to hold the value
	    // together with the null character.
	    // ***************************************
	    char * theseChars = new char[ c -> size() + 1 ];

	    fits_read_col_str(
			      fptr,
			      columnNumber + 1,
			      i + 1,
			      1,
			      c -> size(),
			      "UNDEFINED",
			      &theseChars,
			      &anynul,
			      &status );

	    // **********************************
	    // Copy the buffer to the DAL memory.
	    // **********************************
	    strncpy( c -> stringData(), theseChars, c -> size() );

	    delete[] theseChars;

	    break;
	  }
	default :
	  break;
	} // switch

      delete c;

      fits_check(fptr);
    } // for

  // ******************************
  // Everything is Ok, so return 0.
  // ******************************
  RETURN 0;
}


int MemFitsReaderWriter::readFixedLengthColumnData(fitsfile * fptr,
						   Column * column,
						   long offset,
						   long nread,
						   int &  status ) const
{
    PROC

    int   anynul;
    unsigned long columnNumber = column->columnNumber();

    if( status != 0 )
	ALTRETURNV( 1 );

    // Nothing to read
    if(nread == 0)
	ALTRETURNV(0);


  // ************************************
  // Get a data object to access the data
  // ************************************
  ColumnData *data = column->data();

  errstr << "Going to read column " << column->name() << " dataType="
	 << int( column->dataType() ) << " rows=" << column->rows()
	 << " elements=" << column->elements() << " dimensions="
	 << column->dimensions() << Sas::message( Msg::Core, Msg::Verbose );

  unsigned rowoffset = offset;
  unsigned elementstoread = nread * column->elements();
  unsigned dataoffset = rowoffset * column->elements();

  errstr << "Will read " << nread << " rows starting from " << rowoffset
	 << " that is to say " << elementstoread
	 << " elements will be read starting from element " << dataoffset
	 << Sas::message( Msg::Core, Msg::Verbose );


  switch( column -> dataType() )
  {
      case Column::Bool :
	  
      fits_read_col(
		    fptr,
		    toFITSDataType( column -> dataType() ),
		    columnNumber + 1,
		    1 + rowoffset,
		    1,
		    elementstoread,
		    0,
		    data -> boolData() + dataoffset,
		    &anynul,
		    &status );

      break;

    case Column::Int8 :

      fits_read_col(
		    fptr,
		    toFITSDataType( column -> dataType() ),
		    columnNumber + 1,
		    1 + rowoffset,
		    1,
		    elementstoread,
		    0,
		    data -> int8Data() + dataoffset,
		    &anynul,
		    &status );

      break;

    case Column::Uint16 :

      fits_read_col_usht(
			 fptr,
			 //toFITSDataType( column -> dataType() ),
			 columnNumber + 1,
			 1 + rowoffset,
			 1,
			 elementstoread,
			 0,
			 data -> uint16Data() + dataoffset,
			 &anynul,
			 &status );

      break;

    case Column::Int16 :

      fits_read_col(
		    fptr,
		    toFITSDataType( column -> dataType() ),
		    columnNumber + 1,
		    1 + rowoffset,
		    1,
		    elementstoread,
		    0,
		    data -> int16Data() + dataoffset,
		    &anynul,
		    &status );

      break;

    case Column::Uint32 :

      fits_read_col_uint(
			 fptr,
			 //toFITSDataType( column -> dataType() ),
			 columnNumber + 1,
			 1 + rowoffset,
			 1,
			 elementstoread,
			 0,
			 data -> uint32Data() + dataoffset,
			 &anynul,
			 &status );

      break;

    case Column::Int32 :

      fits_read_col(
		    fptr,
		    toFITSDataType( column -> dataType() ),
		    columnNumber + 1,
		    1 + rowoffset,
		    1,
		    elementstoread,
		    0,
		    data -> int32Data() + dataoffset,
		    &anynul,
		    &status );

      break;

    case Column::Real32 :

      fits_read_col(
		    fptr,
		    toFITSDataType( column -> dataType() ),
		    columnNumber + 1,
		    1 + rowoffset,
		    1,
		    elementstoread,
		    0,
		    data -> real32Data() + dataoffset,
		    &anynul,
		    &status );

      break;

    case Column::Real64 :

      fits_read_col(
		    fptr,
		    toFITSDataType( column -> dataType() ),
		    columnNumber + 1,
		    1 + rowoffset,
		    1,
		    elementstoread,
		    0,
		    data -> real64Data() + dataoffset,
		    &anynul,
		    &status );

      break;

    case Column::DString :
      {
	// ******************************************
	// FITS will null-terminate strings, so it is
	// necessary to allocate a buffer which is
	// large enough to hold the string value
	// together with the null character.
	// ******************************************
	//unsigned long len = column -> size()[ 0 ];
	const unsigned long n = nread;
	const unsigned long len = column -> elements();
	char *   chars = data->stringData() + dataoffset;
	char *   theseChars = new char[ len + 1 ];

	for( unsigned long i = 0; i < n; ++i )
	  {
	    fits_read_col_str(
			      fptr,
			      columnNumber + 1,
			      1 + i + rowoffset,
			      1,
			      1,
			      "UNDEFINED",
			      &theseChars,
			      &anynul,
			      &status );


	    // **********************************
	    // Copy the buffer to the DAL memory.
	    // **********************************
	    char *p = &chars[ i * len];
	    // char *strncpy(char *dest, const char *src, size_t n);
	    strncpy( p, theseChars, len );
// 	    std::cout << "****" << 1 + i + rowoffset << " " << len << " "
// 		      << theseChars << " -- " << p << " -- " << chars << std::endl;

	  } // for

	delete[] theseChars;

	break;
      }
    default :

      errstr << "could not read the column " << column->name() << " from file " << fits_file_name( fptr ) << " since its data type " << int( column->dataType() ) << " is invalid" << Sas::error("internalError");

      break;

    } // switch

  // ************************************
  // Destroy the data object again
  // ************************************
  delete data;

  fits_check(fptr);

  // ******************************
  // Everything is Ok, so return 0.
  // ******************************
  RETURN 0;

}



int MemFitsReaderWriter::readTable( fitsfile *  fptr,
				    DataSet & dataSet,
				    int &  status ) const
{
    PROC

	if( status != 0 )
	{
	    ALTRETURNV( 1 );
	} // if

    TableInfo tableInfo;

    tableInfo._fptr = fptr;

    FitsReaderWriter::readTable( fptr, dataSet, tableInfo, status );

    // ********************
    // Get the column data.
    // ********************
    Table*   table = dataSet.table( dataSet.blocks() - 1 );
    int  columns( table -> columns() );
    //const unsigned int columns( tableInfo._columns.size() );

    // Sort columns by cell type: fixed length or variable length
    std::vector<Column*> fixedlcolumns;
    std::vector<Column*> variablelcolumns;

    for( int i = 0; i < columns; ++i ) {
	if( !columnDataTypeStatus( tableInfo._columns[ i ]._typeCode ) == FitsReaderWriter::Bad ) {
	    Column*  column( table -> column( i ) );
	    if(column->cellType() == Column::Fixed)
		fixedlcolumns.push_back(column);
	    else
		variablelcolumns.push_back(column);
	}
    }

    long nopt = 0;
    fits_get_rowsize(fptr, &nopt, &status);
    unsigned long chunk =  static_cast<unsigned long>(nopt);
    if(chunk == 0 || chunk > table->rows())
	chunk = table->rows();
    errstr << "Optimal number of rows to read for this table: "
	   << chunk << " (nrows = " << table->rows() << ")"
	   << Sas::message( Msg::Core,Msg::Verbose );
    
    unsigned nchunks = chunk == 0 ? 0 : table->rows() / chunk;
    unsigned got = 0;
    for(unsigned K = 0; K <= nchunks; ++K){
      long offset = K*chunk;
      long nread = (K != nchunks) ? chunk : (table->rows() - nchunks*chunk);
      got += nread;
      for(unsigned long k = 0; k < fixedlcolumns.size(); ++k){
	  readFixedLengthColumnData( fptr, fixedlcolumns[k], offset, nread, status );

      }
    }

    for(unsigned long k = 0; k < variablelcolumns.size(); ++k){
	readVariableLengthColumnData( fptr, variablelcolumns[k], status );
    } 
    

    // ******************************
    // Everything is Ok, so return 0.
    // ******************************
    RETURN 0;

}


static int
readArrayData(
	      fitsfile * fptr,
	      Array &  array,
	      int &  status )
{

  PROC


    if( status != 0 )
      {
	ALTRETURNV( 1 );
      } // if

  // *********************************
  // Exit if there is no data to read. 
  // *********************************
  if( numberOfElements( array.size() ) == 0 )
    {
      ALTRETURNV( 0 );
    } // if

  int  anynul;

  // ************************************
  // Get a data object to access the data
  // ************************************
  ArrayData *data = array.data();

  // ******************************************************
  // Read the image. Note that the indices are reverted
  // since FITS uses Fortran indices (NAXIS1 runs fastest).
  // ******************************************************

  switch( array.dataType() )
    {
    case Array::Int8 :

      switch( array.dimensions() )
	{
	case 1 :

	  fits_read_img_byt(
			    fptr,
			    0,
			    1,
			    array.size()[ 0 ],
			    0,
			    (unsigned char*)data -> int8Data(),
			    &anynul,
			    &status );

	  break;

	case 2 :

	  fits_read_2d_byt(
			   fptr,
			   0,
			   0,
			   array.size()[ 1 ],
			   array.size()[ 1 ],
			   array.size()[ 0 ],
			   (unsigned char *)data -> int8Data(),
			   &anynul,
			   &status );

	  break;

	case 3 :

	  fits_read_3d_byt(
			   fptr,
			   0,
			   0,
			   array.size()[ 2 ],
			   array.size()[ 1 ],
			   array.size()[ 2 ],
			   array.size()[ 1 ],
			   array.size()[ 0 ],
			   (unsigned char *)data -> int8Data(),
			   &anynul,
			   &status );

	  break;

	default :

	  errstr << "cannot read array '" << array.name() << "' from file '" << fits_file_name( fptr ) << "' since its dimension " << array.dimensions() << " is not supported" << Sas::error("invalidArrayDimension");

	  break;

	} // switch

      break;

    case Array::Uint16 :

      switch( array.dimensions() )
	{
	case 1 :

	  fits_read_img_usht(
			     fptr,
			     0,
			     1,
			     array.size()[ 0 ],
			     0,
			     data -> uint16Data(),
			     &anynul,
			     &status );

	  break;

	case 2 :

	  fits_read_2d_usht(
			    fptr,
			    0,
			    0,
			    array.size()[ 1 ],
			    array.size()[ 1 ],
			    array.size()[ 0 ],
			    data -> uint16Data(),
			    &anynul,
			    &status );

	  break;

	case 3 :

	  fits_read_3d_usht(
			    fptr,
			    0,
			    0,
			    array.size()[ 2 ],
			    array.size()[ 1 ],
			    array.size()[ 2 ],
			    array.size()[ 1 ],
			    array.size()[ 0 ],
			    data -> uint16Data(),
			    &anynul,
			    &status );

	  break;

	default :

	  errstr << "cannot read array '" << array.name() << "' from file '" << fits_file_name( fptr ) << "' since its dimension " << array.dimensions() << " is not supported" << Sas::error("invalidArrayDimension");

	  break;


	} // switch

      break;

    case Array::Int16 :

      switch( array.dimensions() )
	{
	case 1 :

	  fits_read_img_sht(
			    fptr,
			    0,
			    1,
			    array.size()[ 0 ],
			    0,
			    data -> int16Data(),
			    &anynul,
			    &status );

	  break;

	case 2 :

	  fits_read_2d_sht(
			   fptr,
			   0,
			   0,
			   array.size()[ 1 ],
			   array.size()[ 1 ],
			   array.size()[ 0 ],
			   data -> int16Data(),
			   &anynul,
			   &status );

	  break;

	case 3 :

	  fits_read_3d_sht(
			   fptr,
			   0,
			   0,
			   array.size()[ 2 ],
			   array.size()[ 1 ],
			   array.size()[ 2 ],
			   array.size()[ 1 ],
			   array.size()[ 0 ],
			   data -> int16Data(),
			   &anynul,
			   &status );

	  break;

	default :

	  errstr << "cannot read array '" << array.name() << "' from file '" << fits_file_name( fptr ) << "' since its dimension " << array.dimensions() << " is not supported" << Sas::error("invalidArrayDimension");

	  break;


	} // switch

      break;

    case Array::Uint32 :

      switch( array.dimensions() )
	{
	case 1 :

	  fits_read_img_uint(
			     fptr,
			     0,
			     1,
			     array.size()[ 0 ],
			     0,
			     data -> uint32Data(),
			     &anynul,
			     &status );

	  break;

	case 2 :

	  fits_read_2d_uint(
			    fptr,
			    0,
			    0,
			    array.size()[ 1 ],
			    array.size()[ 1 ],
			    array.size()[ 0 ],
			    data -> uint32Data(),
			    &anynul,
			    &status );

	  break;

	case 3 :

	  fits_read_3d_uint(
			    fptr,
			    0,
			    0,
			    array.size()[ 2 ],
			    array.size()[ 1 ],
			    array.size()[ 2 ],
			    array.size()[ 1 ],
			    array.size()[ 0 ],
			    data -> uint32Data(),
			    &anynul,
			    &status );

	  break;

	default :

	  errstr << "cannot read array '" << array.name() << "' from file '" << fits_file_name( fptr ) << "' since its dimension " << array.dimensions() << " is not supported" << Sas::error("invalidArrayDimension");

	  break;


	} // switch

      break;

    case Array::Int32 :

      switch( array.dimensions() )
	{
	case 1 :

	  fits_read_img_int(
			    fptr,
			    0,
			    1,
			    array.size()[ 0 ],
			    0,
			    data -> int32Data(),
			    &anynul,
			    &status );

	  break;

	case 2 :

	  fits_read_2d_int(
			   fptr,
			   0,
			   0,
			   array.size()[ 1 ],
			   array.size()[ 1 ],
			   array.size()[ 0 ],
			   data -> int32Data(),
			   &anynul,
			   &status );

	  break;

	case 3 :

	  fits_read_3d_int(
			   fptr,
			   0,
			   0,
			   array.size()[ 2 ],
			   array.size()[ 1 ],
			   array.size()[ 2 ],
			   array.size()[ 1 ],
			   array.size()[ 0 ],
			   data -> int32Data(),
			   &anynul,
			   &status );

	  break;

	default :

	  errstr << "cannot read array '" << array.name() << "' from file '" << fits_file_name( fptr ) << "' since its dimension " << array.dimensions() << " is not supported" << Sas::error("invalidArrayDimension");

	  break;


	} // switch

      break;

    case Array::Real32 :

      switch( array.dimensions() )
	{
	case 1 :

	  fits_read_img_flt(
			    fptr,
			    0,
			    1,
			    array.size()[ 0 ],
			    0,
			    data -> real32Data(),
			    &anynul,
			    &status );

	  break;

	case 2 :

	  fits_read_2d_flt(
			   fptr,
			   0,
			   0,
			   array.size()[ 1 ],
			   array.size()[ 1 ],
			   array.size()[ 0 ],
			   data -> real32Data(),
			   &anynul,
			   &status );

	  break;

	case 3 :

	  fits_read_3d_flt(
			   fptr,
			   0,
			   0,
			   array.size()[ 2 ],
			   array.size()[ 1 ],
			   array.size()[ 2 ],
			   array.size()[ 1 ],
			   array.size()[ 0 ],
			   data -> real32Data(),
			   &anynul,
			   &status );

	  break;

	default :

	  errstr << "cannot read array '" << array.name() << "' from file '" << fits_file_name( fptr ) << "' since its dimension " << array.dimensions() << " is not supported" << Sas::error("invalidArrayDimension");

	  break;


	} // switch

      break;

    case Array::Real64 :

      switch( array.dimensions() )
	{
	case 1 :

	  fits_read_img_dbl(
			    fptr,
			    0,
			    1,
			    array.size()[ 0 ],
			    0,
			    data -> real64Data(),
			    &anynul,
			    &status );

	  break;

	case 2 :

	  fits_read_2d_dbl(
			   fptr,
			   0,
			   0,
			   array.size()[ 1 ],
			   array.size()[ 1 ],
			   array.size()[ 0 ],
			   data -> real64Data(),
			   &anynul,
			   &status );

	  break;

	case 3 :

	  fits_read_3d_dbl(
			   fptr,
			   0,
			   0,
			   array.size()[ 2 ],
			   array.size()[ 1 ],
			   array.size()[ 2 ],
			   array.size()[ 1 ],
			   array.size()[ 0 ],
			   data -> real64Data(),
			   &anynul,
			   &status );

	  break;

	default :

	  errstr << "cannot read array '" << array.name() << "' from file '" << fits_file_name( fptr ) << "' since its dimension " << array.dimensions() << " is not supported" << Sas::error("invalidArrayDimension");

	  break;


	} // switch

      break;

    default :

      errstr << "cannot read array '" << array.name() << "' from file '" << fits_file_name( fptr ) << "' since it has an invalid data type " << int( array.dataType() ) << Sas::error("internalError");

      break;

    } // switch

  // ************************************
  // Delete the data object again
  // ************************************
  delete data;

  fits_check(fptr);

  // ******************************
  // Everything is Ok, so return 0.
  // ******************************
  RETURN 0;

}


int
MemFitsReaderWriter::readArray(
			       fitsfile * fptr,
			       DataSet & dataSet,
			       int &  status ) const
{
  PROC

    string   arrayName;
  string   arrayComment;
  int   bitpix;
  Array::DataType  dataType;
  int   naxis;
  long   naxes[ 4 ];
  string   units;
  const unsigned int blockNumber = dataSet.blocks();

  if( status != 0 )
    {
      ALTRETURNV( 1 );
    } // if


  // ===================================================
  // Get the array name. This is achieved by getting the
  // FITS file keyword EXTNAME.
  // ===================================================
  getBlockNameAndComment( fptr, blockNumber, arrayName, arrayComment, status );


  // ********************
  // Get array data type.
  // ********************
  getArrayDataType(
		   fptr,
		   bitpix,
		   dataType,
		   status );

  checkAndSetUnsignedArray(
			   fptr,
			   dataType,
			   status );

  // *********************
  // read the array units.
  // *********************
  readArrayUnits( fptr, units, status );


  // ************************
  // Get the array dimensions
  // ************************
  readArrayDimensions( fptr, naxis, naxes, status );


  // *****************************
  // Add the array to the dataset.
  // *****************************
  Array * array = dataSet.addArray(
				   arrayName,
				   dataType,
				   vector<unsigned long>(
							 (unsigned long *)naxes,
							 (unsigned long *)(naxes + naxis)),
				   arrayComment,
				   units );

  // **********************
  // Read array attributes.
  // **********************
  KeywordChecker * ac = new KeywordChecker;

  //readHeader( fptr, (Attributable &)*array, "", 0, status );
  readHeader( fptr, (Attributable &)*array, ac, status );
  delete ac;

  readArrayScalingFactors(
			  fptr,
			  *array,
			  status );

  // *************************
  // Get the array null value.
  // *************************
  readArrayNullValue(
		     fptr,
		     *array,
		     status );

  readArrayData( fptr, *array, status );

  // ******************************
  // Everything is Ok, so return 0.
  // ******************************
  RETURN 0;

}


int
MemFitsReaderWriter::readPrimaryArray(
				      fitsfile *  fptr,
				      DataSet &  dataSet,
				      KeywordChecker * keywordChecker,
				      int &   status ) const
{
  PROC

    if( status != 0 )
      {
	ALTRETURNV( 1 );
      } // if

  const string arrayComment = "Primary Array";
  int  bitpix;
  char  comment[ 1024 ];
  int  naxis;
  long  naxes[ 4 ];
  string  units;

  // *****************************************
  // Read the dimensions of the primary array.
  // If the naxis( number of dimensions) is
  // zero, then there is no primary image,
  // in which case, just exit.
  // *****************************************
  readArrayDimensions( fptr, naxis, naxes, status );

  if( naxis == 0 )
    {
      ALTRETURNV( 0 );
    } // if

  // *******************
  // read the data type.
  // *******************
  fits_read_key(
		fptr,
		TINT,
		"BITPIX",
		&bitpix,
		comment,
		&status );


  // *********************
  // Read the array units.
  // *********************
  readArrayUnits( fptr, units, status );



  // *****************************
  // Add the array to the dataset.
  // *****************************
  Array * array = dataSet.addArray(
				   nameOfPrimaryImage,
				   fromFITSBitpix( bitpix ),
				   vector<unsigned long>(
							 (unsigned long *)naxes,
							 (unsigned long *)(naxes + naxis)),
				   arrayComment,
				   units );


  // **********************
  // Read array attributes.
  // **********************
  readHeader( fptr, (Attributable &)*array, keywordChecker, status );


  // *********************
  // Read scaling factors.
  // *********************
  readArrayScalingFactors(
			  fptr,
			  *array,
			  status );


  // ****************
  // Read array data.
  // ****************
  readArrayData( fptr, *array, status );


  // ******************************
  // Everything is Ok, so return 0.
  // ******************************
  RETURN 0;
}


static void
deleteCopyrightMessage( fitsfile * fptr,
			int *  status )
{
  PROC

    if( *status != 0 )
      {
	ENDPROC
	  } // if

  // **********************************************
  // Delete the copyright message.
  // This will be achieved by deleting all comments
  // in the (newly created) primary header unit.
  // Just keep deleting the COMMENT keywords, until
  // there are no more (indicated by KEY_NO_EXIST).
  // **********************************************
  //int   hdutype;

  //fits_movabs_hdu( fptr, 1, &hdutype, status );
  //fits_delete_record( fptr, 5, status );
  //fits_delete_record( fptr, 6, status );
  //fits_delete_record( fptr, 7, status );
  //fits_delete_record( fptr, 8, status );
  while( *status == 0 ) fits_delete_key( fptr, "COMMENT*", status );
  //fits_delete_key( fptr, "COMMENT*", status );

  if( *status != KEY_NO_EXIST ) fits_check(fptr);

  *status = 0;

  ENDPROC
    }


//MemFitsReaderWriter::MemFitsReaderWriter()
//{
// PROC
// ENDPROC
//}


//MemFitsReaderWriter::~MemFitsReaderWriter()
//{
// PROC
// ENDPROC
//}



void
MemFitsReaderWriter::write( DataSet* dataSet )
{

  PROC

    _dataSet = dataSet;

  const unsigned int numberOfBlocks = dataSet -> blocks();
  const string  setName = dataSet -> name();

  dataSetServer -> clobberDataSet( setName );

  //const bool  exists = dataSetServer -> exists( fileName );

  // ****************************************************
  // Check to see if the dataset may be created.
  // If it may, then attempt to clobber it, if necessary.
  // ****************************************************
  //if( exists && !clobber() )
  //{
  // errstr << "The dataset with name '" << fileName << "' could not be created because it already exists and clobber is not set." << Sas::error( "datasetAlreadyExists" );

  //}
  //else if( exists )
  //{
  // ====================
  // Clobber the dataset.
  // ====================
  //status = unlink( fileName.c_str() ); // UNIX delete file.

  //if( status != 0 )
  //{
  // errstr << "The dataset with name '" << fileName << "' could not be created because it already exists and clobber is set. but the clobber failed" << Sas::error("clobberFailed");
  //} // if
  //}
  //else if( clobber() )
  //{
  // ===================================
  // Clobber the file , if there is one.
  // ===================================
  //unlink( fileName.c_str() ); // UNIX delete file.
  //} // if

  // ******************************
  // Attempt to create the dataset.
  // ******************************
  fitsCreateFile( &_fptr, setName.c_str(), &status );

  if( status )
    {

      errstr << "The dataset with name '" << setName << "' could not be created." << Sas::error("createDatasetFailed");

    } // if

  fits_check( _fptr );

  // ******************************************
  // If the number of blocks is zero, or if the
  // the first block is a table, or if
  // the first block is an array which does
  // not have name nameOfPrimaryImage, then create
  // a (FITS required) dummy primary image.
  // ******************************************
  if(
     _dataSet -> blocks() == 0
     ||
     _dataSet -> block( 0 ) -> type() == Block::TableT
     ||
     _dataSet -> block( 0 ) -> type() == Block::ArrayT
     &&
     !(_dataSet -> block( 0 ) -> name() == nameOfPrimaryImage) )
    {
      fits_create_img( _fptr, 8, 0, 0, &status );
      fits_check(_fptr);
      // ==================================================
      // Delete the copyright message from the fits header.
      // ==================================================
      deleteCopyrightMessage( _fptr, &status );

      // **************************
      // Write the process history.
      // **************************
      writeProcessHistory( _fptr, _dataSet, status );

      // ********************
      // Write Debug history.
      // ********************
      writeDebugHistory( _fptr, _dataSet, status );

      // ***************************
      // Write the Creator and Date.
      // ***************************
      writeCreatorAndDate( _fptr, _dataSet, status );

      // ===========================================
      // write the DataSet attributes  and separator
      // to the FITS primary header.
      // ===========================================
      writeHeader( _fptr, *_dataSet, "", 0, status );
    } // if


  // *******************************
  // For each Block in the Data Set:
  // *******************************
  for( unsigned int i = 0; i < numberOfBlocks; i++ )
    {
      const Block *b = _dataSet -> block(i);

      switch( b->type() )
	{
	case Block::TableT :
	  {
	    Table *table = _dataSet -> table( i );
	    table -> seek( 0, table -> rows() );
	    writeTable( _fptr, *table, status );
	  }
	break;

	case Block::ArrayT :
	  {
	    const Array * array = _dataSet -> array( i );

	    writeArray( _fptr, *array, status );

	    if( i == 0 && array -> name() == nameOfPrimaryImage )
	      {
		// ==================================================
		// Delete the copyright message from the fits header.
		// ==================================================
		deleteCopyrightMessage( _fptr, &status );

		// **************************
		// Write the process history.
		// **************************
		writeProcessHistory( _fptr, _dataSet, status );

		// ********************
		// Write Debug history.
		// ********************
		writeDebugHistory( _fptr, _dataSet, status );

		// ***************************
		// Write the Creator and Date.
		// ***************************
		writeCreatorAndDate( _fptr, _dataSet, status );

		// ===========================================
		// write the DataSet attributes  and separator
		// to the FITS primary header.
		// ===========================================
		writeHeader( _fptr, *_dataSet, "", 0, status );
		writeSeparator( _fptr, array -> attributes(), status );
	      }
	    else
	      {
		// ===================================================
		// Set the array name. This is achieved by setting the
		// FITS file keyword EXTNAME.
		// ===================================================
		fits_write_key_str(
				   _fptr,
				   "EXTNAME", // This is how to set the name.
				   (char *)array->name().c_str(),
				   array->label()=="" ? (char*)defaultArrayComment.c_str() : (char *)array->label().c_str(),
				   &status );
		fits_check(_fptr);
	      } // if

	    // ===========================
	    // Write the array attributes.
	    // ===========================
	    writeHeader( _fptr, (const Attributable&)*array, "", 0, status );
	  }
	break;

	default:
	  errstr<<"could not write data set '" << _dataSet->name() << "' since an invalid block type " << int( b->type() ) << " was found for block " << b->name() << Sas::error("internalError");

	} // switch
    } // for


  writeLongStringWarning( _fptr, status );


  // ****************
  // Close FITS file.
  // ****************
  fitsCloseFile( _fptr, &status );

  fits_check(_fptr);

  ENDPROC

    }


void
MemFitsReaderWriter::read( DataSet* dataSet )
{

  PROC

    _dataSet = dataSet;

  int   hdutype;
  char   value[ 1024 ];
  const string  fileName = _dataSet -> name();

  // ******************************
  // Check that the dataset exists.
  // ******************************
  if( !dataSetServer -> exists( fileName ) )
    {
      errstr << "The dataset with name '" << fileName << "' could not be read; either it does not exist or it is in an unrecognised format" << Sas::error( "datasetCouldNotBeRead" );
    } // if

  // **********************************************
  // Open the FITS Data File.
  // Then check that the file opened without error.
  // **********************************************
  fitsOpenFile(&_fptr, fileName.c_str(), READONLY, &status );
  fits_check(_fptr);


  // **********************
  // Move to primary array.
  // **********************
  int hdunum = 1;   // Primary header in position 1.

  fits_movabs_hdu( _fptr, hdunum, &hdutype, &status );
  fits_check(_fptr);

  // *********************
  // Read process history.
  // *********************
  readProcessHistory( _fptr, _dataSet, status );

  // *******************
  // Read Debug history.
  // *******************
  readDebugHistory( _fptr, _dataSet, status );

  // ***************************
  // Read the Creator and Date.
  // ***************************
  readCreatorAndDate( _fptr, _dataSet, status );

  // *************************************
  // Read primary array (if there is one).
  // *************************************
  PrimaryKeywordChecker * pkc = new PrimaryKeywordChecker();
  readPrimaryArray( _fptr, *_dataSet, pkc, status );
  delete pkc;

  // *****************************************************
  // read the data set attributes from the primary header.
  // *****************************************************
  DataSetKeywordChecker * dskc = new DataSetKeywordChecker();
  readHeader( _fptr, *_dataSet, dskc, status );
  delete dskc;

  // ********************************
  // For each block in the FITS file.
  // ********************************

  // =======================
  // Begin with first block.
  // =======================
  hdunum = 2;   // First table in first extended unit.

  fits_movabs_hdu( _fptr, hdunum, &hdutype, &status );

  if (status && status != END_OF_FILE) fits_check( _fptr );

  // ===============
  // For each block.
  // ===============
  while( !status )
    {

      // ======================================
      // Check the block type. This is achieved
      // by getting the XTENSION (which is a
      // a mandatory keyword) keyword value:
      // BINTABLE - a binary table
      //
      // ======================================
      fits_read_keyword(
			_fptr,
			"XTENSION",
			value,
			(char *)0,
			&status );

      string cleanValue = trim(unquote(value));

      if(
	 ( cleanValue == "BINTABLE" )
	 ||
	 ( cleanValue ==  "ASCII" ) )
	{
	  readTable( _fptr, *_dataSet, status );

	}
      else if( cleanValue == "IMAGE" )
	{
	  readArray( _fptr, *_dataSet, status );

	}
      else
	{
	  errstr << "Unrecognized XTENSION '" << value << "'"
		 << Sas::error( "FITSFailure" );

	} // if

      // =======================
      // Look for another block.
      // =======================
      hdunum++;

      fits_movabs_hdu( _fptr, hdunum, &hdutype, &status );

      if (status && status != END_OF_FILE) fits_check(_fptr);
    } // while


  // ****************
  // Close FITS file.
  // ****************
  status = 0;

  fitsCloseFile( _fptr, &status );

  fits_check(_fptr);

  ENDPROC;

}


/// OLD CODE

/*
int
MemFitsReaderWriter::readColumn(
 fitsfile * fptr,
 int &  columnNumber,
 Table &  table,
 int &  status ) const
{
 PROC

 if( status != 0 )
 {
  ALTRETURNV( 1 );
 } // if

 char   name[ 32 ];
 int   typeCode;
 char   comment[ 1024 ];
 char   columnComment[ 1024 ];
 char   units[ 32 ];
 string   tableName;
 const unsigned int maxdim = 4;
 int   naxis;
 long   naxes[ maxdim ];
 string   type = "TTYPE";
 Column::DataType  dataType;
 Column *   column;


 // *************************************
 // Get the column name & column comment.
 // *************************************
 fits_read_key(
   fptr,
   TSTRING,
   (char *)( type + toString( columnNumber + 1 )).c_str(),
   name,
   columnComment,
   &status );


 // *******************************
 // Read the optional column units.
 // *******************************
 readColumnUnits( fptr, columnNumber, units, comment, status );


 // ***************
 // Get dimensions.
 // ***************
 fits_read_tdim(
    fptr,
    columnNumber + 1,
    maxdim,
    &naxis,
    naxes,
    &status );

 fits_check(fptr);

 // *************************
 // Get the column data type.
 // *************************
 getColumnDataType(
  fptr,
  columnNumber,
  typeCode,
  dataType,
  status );

 if( !isValidDalDataType( abs( typeCode )) )
 {
  errstr << "The column with column number " << columnNumber << " has a fits typecode " << typeCode << " which is not supported by the DAL" << Sas::error( "unsupportedColumnType" );
 } // if

 checkAndSetUnsignedColumn(
   fptr,
   columnNumber,
   dataType,
   status );

 // ********************************
 // Now add the column to the table.
 // ********************************

 if( typeCode >= 0 )
 {
  column = table.addColumn(
     name,
     dataType,
     columnComment,
     units,
     vector<unsigned long>(
      (unsigned long *)(naxes),
      (unsigned long *)(naxes + naxis) ));
 }
 else
 {
  column = table.addColumn(
     name,
     dataType,
     columnComment,
     units,
     Dal::zero );

 }// if


 errstr << "added column " << column->name() << " to table " << table.name() << " type=" << typeCode << " dataType=" << int( dataType ) << " rows=" << column->rows() << " elements=" << column->elements() << " dimensions=" << column->dimensions() << Sas::message( Msg::Core, Msg::Verbose );

 if( !getenv( "SAS_NO_COLUMN_ATTRIBUTES" ) )
 {
  // ***********************
  // Read Column Attributes.
  // ***********************
  ColumnKeywordChecker * cac = new ColumnKeywordChecker( column );

  readAttributes(
    fptr,
    (Attributable &)*column,
    cac,
    status );

  delete cac;

 } // if

 // ***********************
 // TEMPORARILY DISABLED:
 // Get the Scaling Factors
 // ***********************
 //readColumnScalingFactors(
 // fptr,
 // columnNumber,
 // *column,
 // status );

 // **************************
 // Get the column null value.
 // **************************
 readColumnNullValue(
  fptr,
  *column,
  status );

 // ********************
 // Get the column data.
 // ********************
 if( typeCode >= 0 )
 {
  readFixedLengthColumnData( fptr, column, columnNumber, status );
 }
 else
 {
  readVariableLengthColumnData( fptr, column, columnNumber, status );
 } // if


 // ******************************
 // Everything is Ok, so return 0.
 // ******************************
 RETURN 0;

}
*/

/*
int
MemFitsReaderWriter::readTable(
 fitsfile *  fptr,
 int &  blockNumber,
 DataSet & dataSet,
 int &  status ) const
{
 PROC

 long   rows;
 char    card[ 1024 ];
 char    comment[ 1024 ];
 int   columns;
 string   tableName;
 string   tableComment;

 if( status != 0 )
 {
  ALTRETURNV( 1 );
 } // if


 // ====================
 // Get the column size.
 // ====================
 fits_read_key_lng(
   fptr,
   "NAXIS2",
   &rows,
   comment,
   &status );

 fits_check(fptr);


 // ===================================================
 // Get the table name. This is achieved by getting the
 // FITS file keyword EXTNAME.
 // ===================================================
 getBlockNameAndComment( fptr, blockNumber, tableName, tableComment, status );

 // If the name already existed, append the block number
 if ( dataSet.hasBlock(tableName) )
  tableName += toString(blockNumber++);


 // ******************************
 // Add the table to the data set.
 // ******************************
 Table * table = dataSet.addTable( tableName, rows, tableComment );

 errstr << "added table " << Sas::message( Msg::Core, Msg::Verbose );

 // ======================
 // read the table header.
 // ======================
 TableKeywordChecker * tac = new TableKeywordChecker( table );
 readHeader( fptr, (Attributable &)*table, tac, status );
 delete tac;

 // ============================================
 // Get the number of columns in the FITS table.
 // ============================================

 // -----------------------------------------------
 // Ensure that search is from top of table header.
 // -----------------------------------------------
 fits_read_record( fptr, 0, card, &status );

 fits_read_key(
   fptr,
   TINT,
   "TFIELDS", // This is how to get no. cols.
   &columns,
   comment,
   &status );

 fits_check(fptr);

 // *****************
 // read each Column.
 // *****************
 for( int i = 0; i < columns; i++ )
 {
  readColumn(
    fptr,
    i,
    *table,
    status );

 } // for


 // ******************************
 // Everything is Ok, so return 0.
 // ******************************
 RETURN 0;

}
*/
