using namespace std;
/*************************************************************** 
* 
* multidimcolumn.cc & XMM Science Analysis System (SAS) 
* 
* DAL Test harness which tests multi-dimensional columns.
* 
* (c) 1999 ESA - Written by Mark Thomas, CARA
* 
* $Id: multidimcolumn.cc,v 1.2 2004/03/05 14:43:03 eojero Exp $ 
* 
*****************************************************************/ 
#include<cassert>

#include <Dal.h>
#include <cmath>
#include "dal_utilities.h"
#include "Tracer.h"
#include "StopWatch.h"
#include "stringutil.h"


template<class T>
bool feq(T v1, T v2) { PROC RETURN abs(v1-v2) < 1e-5; }


template<class T, int a, int b>
T numericValue( unsigned long row, unsigned long element )
{
  return row + T(a) / T(b) * element;
}


template<class T>
static void
fill(
	const string&			setName,
	const string&			tableName,
	const string&			columnName,
	Column::DataType		dataType,
	unsigned int			numberOfColumns,
	unsigned long 			nRows,
	const vector<unsigned long>&	size,
	T (*value)( unsigned long row, unsigned long elements ),
	T * (Data::*getdata)() const )
{

  DataSet* set = dataSetServer->open(setName,DataSet::Create);
  assert( set -> name() == setName );
  Table* tab = set->addTable(tableName,nRows);
  assert( tab->rows() == nRows );
  assert(tab -> name() == tableName );

  Column *col;

  for( unsigned int i = 0; i < numberOfColumns; ++i )
  {
    col = tab->addColumn(columnName + toString( i ),dataType,"","",size);
    assert( tab -> columns() == i + 1 );
    assert( col -> name() == columnName + toString( i ) );
    assert( col -> dataType() == dataType );
    assert(col->size() == size);
    assert(col->elements() == numberOfElements( size ));

    ColumnData * d = col->data();

    assert(d->elements() == numberOfElements( size ));
    assert(d->size() == size);
    assert(d->count() == nRows);
    assert(d->from() == 0);

    T * data = (((Data*)(d)) ->* getdata)();

    for (unsigned long row = d->from(); row < d->from()+d->count(); ++row)
    {
      for (unsigned long element = 0; element < d->elements(); ++element)
      {
        unsigned long pos = d->position(row,element);
        assert( pos == row * numberOfElements( size ) + element );
        data[d->position(row,element)] = value(row,element);
      } // for
    } // for
    delete d;
  } // for

  dataSetServer->close(set);

}


template<class T>
static void
checkAll(
	const string&			setName,
	const string&			tableName,
	const string&			columnName,
	Column::DataType		dataType,
	unsigned long			numberOfColumns,
	unsigned long 			nRows,
	const vector<unsigned long>&	size,
	T (*value)( unsigned long row, unsigned long elements ),
	T * (Data::*getdata)() const )
{

  DataSet* set = dataSetServer->open(setName,DataSet::Read);
  assert( set -> name() == setName );
  Table* tab = set->table(tableName);
  assert( tab->rows() == nRows );
  assert(tab -> name() == tableName );
  assert( tab -> columns() == numberOfColumns ); 

  Column **cols = new (Column*)[ numberOfColumns ];
  ColumnData ** ds = new (ColumnData*)[ numberOfColumns ];
  T ** datas = new (T*)[ numberOfColumns ];

  for( unsigned int i = 0; i < numberOfColumns; ++i )
  {
    cols[ i ] = tab->column( columnName + toString( i ) );
    assert( cols[ i ] -> name() == columnName + toString( i ) );
    assert( cols[ i ] -> dataType() == dataType );
    assert( cols[ i ] ->size() == size );
    assert( cols[ i ] ->elements() == numberOfElements( size ));

    ds[ i ] = cols[ i ] -> data();

    assert( ds[ i ] -> elements() == numberOfElements( size ));
    assert( ds[ i ] -> size() == size );
    assert( ds[ i ] -> count() == nRows );
    assert( ds[ i ] -> from() == 0 );
    datas[ i ] = (((Data *)(ds[i])) ->* getdata)();
  } // for

  for( unsigned int i = 0; i < numberOfColumns; ++i )
  {
    ColumnData ** drs = new (ColumnData*)[ ds[i] -> count() ];
    T ** datars = new (T*)[ ds[i] -> count() ];

    for (unsigned long row = ds[i]->from(); row < ds[i]->from()+ds[i]->count(); ++row)
    {
      drs[ row ] = cols[ i ] -> data( row, 1 );
      datars[ row ] = (((Data *)(drs[row])) ->* getdata)();
    } // for

    T * data = datas[ i ];

    for (unsigned long row = ds[i]->from(); row < ds[i]->from()+ds[i]->count(); ++row)
    {
      T * datar = datars[ row ];

      for (unsigned long element = 0; element < ds[i]->elements(); ++element)
      {
        unsigned long pos = ds[i]->position(row,element);
        assert( pos == row * numberOfElements( size ) + element );
        assert( feq( data[pos], value(row,element ) ));
        assert( feq( datar[element], value(row,element) ));
      } // for
    } // for

    for (unsigned long row = ds[i]->from(); row < ds[i]->from()+ds[i]->count(); ++row)
    {
    	delete drs[ row ];
    } // for
    delete[] drs;
    delete[] datars;
  } // for

  for( unsigned int i = 0; i < numberOfColumns; ++i )
  {
  	delete ds[ i ];
  } // for

  delete[] cols;
  delete[] ds;
  delete[] datas;

  dataSetServer->close(set);

}


template<class T>
static void
check(
	const string&			setName,
	const string&			tableName,
	const string&			columnName,
	Column::DataType		dataType,
	unsigned long			numberOfColumns,
	unsigned long 			nRows,
	const vector<unsigned long>&	size,
	T (*value)( unsigned long row, unsigned long elements ),
	T * (Data::*getdata)() const )
{

  DataSet* set = dataSetServer->open(setName,DataSet::Read);
  assert( set -> name() == setName );
  Table* tab = set->table(tableName);
  assert( tab->rows() == nRows );
  assert(tab -> name() == tableName );
  assert( tab -> columns() == numberOfColumns ); 
  Column *col;

  for( unsigned int i = 0; i < numberOfColumns; ++i )
  {
    col = tab->column( columnName + toString( i ) );
    assert( col -> name() == columnName + toString( i ) );
    assert( col -> dataType() == dataType );
    assert( col->size() == size );
    assert( col->elements() == numberOfElements( size ));

    ColumnData * d = col->data();

    assert(d->elements() == numberOfElements( size ));
    assert( d->size() == size );
    assert( d->count() == nRows );
    assert(d->from() == 0);

    T * data = (((Data *)(d)) ->* getdata)();

    for (unsigned long row = d->from(); row < d->from()+d->count(); ++row)
    {
      ColumnData * dr = col->data(row,1);
      T * datar = (((Data *)(dr)) ->* getdata)();
      for (unsigned long element = 0; element < d->elements(); ++element)
      {
        unsigned long pos = d->position(row,element);
        assert( pos == row * numberOfElements( size ) + element );
        assert( feq( data[pos], value(row,element )) );
        assert( feq( datar[element], value(row,element)) );
      } // for
      delete dr;
    } // for
    delete d;
  } // for

  dataSetServer->close(set);

}


template<class T>
static void
testNumberOfColumns(
	Column::DataType		dataType,
	unsigned int			numberOfColumns,
	unsigned long 			nRows,
	const vector<unsigned long>&	size,
	T (*value)( unsigned long row, unsigned long elements ),
	T * (Data::*getdata)() const )
{
  fill( "test.dat", "table", "column", dataType, numberOfColumns, nRows, size, value, getdata );
  check( "test.dat", "table", "column", dataType, numberOfColumns, nRows, size, value, getdata );
  checkAll( "test.dat", "table", "column", dataType, numberOfColumns, nRows, size, value, getdata );
}


template<class T>
static void
testDataType(
	Column::DataType		dataType,
	unsigned long 			nRows,
	const vector<unsigned long>&	size,
	T (*value)( unsigned long row, unsigned long elements ),
	T * (Data::*getdata)() const )
{
  testNumberOfColumns( dataType, 1, nRows, size, value, getdata );
  testNumberOfColumns( dataType, 2, nRows, size, value, getdata );
  testNumberOfColumns( dataType, 10, nRows, size, value, getdata );
}


static void
testSize( unsigned long nRows, const vector<unsigned long>& size )
{
  //testDataType( Column::Bool, nRows, size, true, &ColumnData::boolData );
  int8 (*int8Value)(unsigned long,unsigned long) = numericValue<int8,1,1>;
  testDataType( Column::Int8, nRows, size, int8Value, &ColumnData::int8Data );
  testDataType( Column::Int16, nRows, size, &numericValue<int16,1,1>, &ColumnData::int16Data );
  testDataType( Column::Int32, nRows, size, &numericValue<int32,1,1>, &ColumnData::int32Data );
  testDataType( Column::Real32, nRows, size, &numericValue<real32,1,10>, &ColumnData::real32Data );
  testDataType( Column::Real64, nRows, size, &numericValue<real64,1,10>, &ColumnData::real64Data );

}


static void
doTest( unsigned long nRows )
{
  vector<unsigned long> size;

  size.push_back(4);
  testSize( nRows, size );

  size.push_back(3);
  testSize( nRows, size );

  size.push_back(2);
  testSize( nRows, size );

  size.push_back(2);
  testSize( nRows, size );

}

int
main()
{
   	StopWatch	sw;
	// doTest( 0 ); Does not work; nRows defaults to 1.
	doTest( 1 );
	doTest( 5 );
	doTest( 10 );
	doTest( 19 );
	//doTest( 101 );
	//doTest( 5000 );
   	cout << "CPU:" << sw.elapsed() << endl;
}


